"""Functions for MusPy objects.

This module defines functions that can be applied to a MusPy object.

Functions
---------

- adjust_resolution
- adjust_time
- append
- clip
- get_end_time
- get_real_end_time
- remove_duplicate
- sort
- to_ordered_dict
- transpos

"""
from collections import OrderedDict
from typing import Callable, Union

from .base import Base, ComplexBase
from .classes import Note, Track
from .music import Music

__all__ = [
    "adjust_resolution",
    "adjust_time",
    "append",
    "clip",
    "get_end_time",
    "get_real_end_time",
    "remove_duplicate",
    "sort",
    "to_ordered_dict",
    "transpose",
]


def adjust_resolution(
    music: Music,
    target: int = None,
    factor: float = None,
    rounding: Union[str, Callable] = "round",
) -> Music:
    """Adjust resolution and timing of all time-stamped objects.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Object to adjust the resolution.
    target : int, optional
        Target resolution.
    factor : int or float, optional
        Factor used to adjust the resolution based on the formula:
        `new_resolution = old_resolution * factor`. For example, a
        factor of 2 double the resolution, and a factor of 0.5 halve the
        resolution.
    rounding : {'round', 'ceil', 'floor'} or callable, default: 'round'
            Rounding mode.

    """
    return music.adjust_resolution(
        target=target, factor=factor, rounding=rounding
    )


def adjust_time(obj: Base, func: Callable[[int], int]) -> Base:
    """Adjust the timing of time-stamped objects.

    Parameters
    ----------
    obj : :class:`muspy.Music` or :class:`muspy.Track`
        Object to adjust the timing.
    func : callable
        The function used to compute the new timing from the old timing,
        i.e., `new_time = func(old_time)`.

    See Also
    --------
    :func:`muspy.adjust_resolution` :
        Adjust the resolution and the timing of time-stamped objects.

    Note
    ----
    The resolution are left unchanged.

    """
    return obj.adjust_time(func=func)


def append(obj1: ComplexBase, obj2) -> ComplexBase:
    """Append an object to the correseponding list.

    This will automatically determine the list attributes to append
    based on the type of the object.

    Parameters
    ----------
    obj1 : :class:`muspy.ComplexBase`
        Object to which `obj2` to append.
    obj2
        Object to be appended to `obj1`.

    Notes
    -----
    - If `obj1` is of type :class:`muspy.Music`, `obj2` can be
      :class:`muspy.Tempo`, :class:`muspy.KeySignature`,
      :class:`muspy.TimeSignature`, :class:`muspy.Lyric`,
      :class:`muspy.Annotation` or :class:`muspy.Track`.
    - If `obj1` is of type :class:`muspy.Track`, `obj2` can be
      :class:`muspy.Note`, :class:`muspy.Chord`,
      :class:`muspy.Lyric` or :class:`muspy.Annotation`.

    See Also
    --------
    :class:`muspy.ComplexBase.append` : Equivalent function.

    """
    return obj1.append(obj2)


def extend(obj1: ComplexBase, obj2, deepcopy: bool = False) -> ComplexBase:
    """Extend the list(s) with another object or iterable.

    Parameters
    ----------
    obj1 : :class:`muspy.ComplexBase`
        Object to extend.
    obj2
        If an object of the same type as `obj1` is given, extend the
        list attributes with the corresponding list attributes of
        `obj2`. If an iterable is given, call `obj1.append` on each
        item.
    deepcopy : bool, default: False
        Whether to make deep copies of the appended objects.

    See Also
    --------
    :class:`muspy.ComplexBase.extend` : Equivalent function.

    """
    return obj1.extend(obj2, deepcopy=deepcopy)


def clip(
    obj: Union[Music, Track, Note], lower: int = 0, upper: int = 127,
) -> Union[Music, Track, Note]:
    """Clip the velocity of each note.

    Parameters
    ----------
    obj : :class:`muspy.Music`, :class:`muspy.Track` or \
            :class:`muspy.Note`
        Object to clip.
    lower : int or float, default: 0
        Lower bound.
    upper : int or float, default: 127
        Upper bound.

    """
    return obj.clip(lower=lower, upper=upper)


def get_end_time(obj: Union[Music, Track], is_sorted: bool = False) -> int:
    """Return the the time of the last event in all tracks.

    This includes tempos, key signatures, time signatures, note offsets,
    lyrics and annotations.

    Parameters
    ----------
    obj : :class:`muspy.Music` or :class:`muspy.Track`
        Object to inspect.
    is_sorted : bool, default: False
        Whether all the list attributes are sorted.

    """
    return obj.get_end_time(is_sorted=is_sorted)


def get_real_end_time(music: Music, is_sorted: bool = False) -> float:
    """Return the end time in realtime.

    This includes tempos, key signatures, time signatures, note offsets,
    lyrics and annotations. Assume 120 qpm (quarter notes per minute) if
    no tempo information is available.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Object to inspect.
    is_sorted : bool, default: False
        Whether all the list attributes are sorted.

    """
    return music.get_real_end_time(is_sorted=is_sorted)


def remove_duplicate(obj: ComplexBase) -> ComplexBase:
    """Remove duplicate change events.

    Parameters
    ----------
    obj : :class:`muspy.Music`
        Object to process.

    """
    return obj.remove_duplicate()


def sort(obj: ComplexBase) -> ComplexBase:
    """Sort all the time-stamped objects with respect to event time.

    - If a :class:`muspy.Music` is given, this will sort key signatures,
      time signatures, lyrics and annotations, along with notes, lyrics
      and annotations for each track.
    - If a :class:`muspy.Track` is given, this will sort notes, lyrics
      and annotations.

    Parameters
    ----------
    obj : :class:`muspy.ComplexBase`
        Object to sort.

    """
    return obj.sort()


def to_ordered_dict(
    obj: Base, skip_missing: bool = True, deepcopy: bool = True
) -> OrderedDict:
    """Return an OrderedDict converted from a Music object.

    Parameters
    ----------
    obj : :class:`muspy.Base`
        Object to convert.
    skip_missing : bool, default: True
        Whether to skip attributes with value None or those that are
        empty lists.
    deepcopy : bool, default: True
        Whether to make deep copies of the attributes.

    Returns
    -------
    OrderedDict
        Converted OrderedDict.

    """
    return obj.to_ordered_dict(skip_missing=skip_missing, deepcopy=deepcopy)


def transpose(
    obj: Union[Music, Track, Note], semitone: int
) -> Union[Music, Track, Note]:
    """Transpose all the notes by a number of semitones.

    Parameters
    ----------
    obj : :class:`muspy.Music`, :class:`muspy.Track` or \
            :class:`muspy.Note`
        Object to transpose.
    semitone : int
        Number of semitones to transpose the notes. A positive value
        raises the pitches, while a negative value lowers the pitches.

    """
    return obj.transpose(semitone=semitone)



"""Utility functions."""
from collections import OrderedDict
from typing import Dict, List, Tuple

import yaml

NOTE_MAP: Dict[str, int] = {
    "C": 0,
    "D": 2,
    "E": 4,
    "F": 5,
    "G": 7,
    "A": 9,
    "B": 11,
}

CIRCLE_OF_FIFTHS: List[Tuple[int, str]] = [
    (4, "Fb"),
    (11, "Cb"),
    (6, "Gb"),
    (1, "Db"),
    (8, "Ab"),
    (3, "Eb"),
    (10, "Bb"),
    (5, "F"),  # Lydian
    (0, "C"),  # Major/Ionian
    (7, "G"),  # Mixolydian
    (2, "D"),  # Dorian
    (9, "A"),  # Minor/Aeolian
    (4, "E"),  # Phrygian
    (11, "B"),  # Locrian
    (6, "F#"),
    (1, "C#"),
    (8, "G#"),
    (3, "D#"),
    (10, "A#"),
    (5, "E#"),
    (0, "B#"),
]

MODE_CENTERS = {
    "major": 8,
    "minor": 11,
    "lydian": 7,
    "ionian": 8,
    "mixolydian": 9,
    "dorian": 10,
    "aeolian": 11,
    "phrygian": 12,
    "locrian": 13,
}

NOTE_TYPE_MAP: Dict[str, float] = {
    "1024th": 0.00390625,
    "512th": 0.0078125,
    "256th": 0.015625,
    "128th": 0.03125,
    "64th": 0.0625,
    "32nd": 0.125,
    "16th": 0.25,
    "eighth": 0.5,
    "quarter": 1.0,
    "half": 2.0,
    "whole": 4.0,
    "breve": 8.0,
    "long": 16.0,
    "maxima": 32.0,
}


def note_str_to_note_num(note_str: str):
    """Return the note number of a note string.

    The regular expression for the note string is `[A-G][#b]*`. The base
    note must be capitalized. There can be multiple accidentals, where
    '#' denotes a sharp and 'b' denotes a flat. Some examples include
    'C'->0, 'D#'->3, 'Eb'->3.

    Parameters
    ----------
    note_str : str
        Note string.

    Returns
    -------
    int, 0-11
        Note number.

    """
    note_num = NOTE_MAP.get(note_str[0])
    if note_num is None:
        raise ValueError(
            f"Expect a base note from 'A' to 'G', but got :{note_str[0]}."
        )
    for alter in note_str[1:]:
        if alter == "#":
            note_num += 1
        elif alter == "b":
            note_num -= 1
        else:
            raise ValueError(
                f"Expect an accidental of '#' or 'b', but got : {alter}."
            )
    if note_num > 11 or note_num < 0:
        return note_num % 12
    return note_num


class OrderedDumper(yaml.SafeDumper):
    """A dumper that supports OrderedDict."""


def _dict_representer(dumper, data):
    return dumper.represent_mapping(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items()
    )


OrderedDumper.add_representer(OrderedDict, _dict_representer)


def yaml_dump(
    data, Dumper=None, allow_unicode: bool = True, **kwargs
):  # pylint: disable=invalid-name
    """Dump data to YAML, which supports OrderedDict.

    Code adapted from https://stackoverflow.com/a/21912744.
    """
    if Dumper is None:
        Dumper = OrderedDumper
    return yaml.dump(
        data, Dumper=Dumper, allow_unicode=allow_unicode, **kwargs
    )



"""Lakh MIDI Dataset."""
from pathlib import Path
from typing import Union

from ..inputs import read_midi
from ..music import Music
from .base import DatasetInfo, RemoteFolderDataset

# pylint: disable=line-too-long

_NAME = "Lakh MIDI Dataset"
_DESCRIPTION = """\
The Lakh MIDI dataset is a collection of 176,581 unique MIDI files, 45,129 of \
which have been matched and aligned to entries in the Million Song Dataset. \
Its goal is to facilitate large-scale music information retrieval, both \
symbolic (using the MIDI files alone) and audio content-based (using \
information extracted from the MIDI files as annotations for the matched \
audio files)."""
_HOMEPAGE = "https://colinraffel.com/projects/lmd/"
_LICENSE = "Creative Commons Attribution 4.0 International License (CC-By 4.0)"
_CITATION = """\
@phdthesis{raffel2016learning,
  author={Colin Raffel},
  title={Learning-Based Methods for Comparing Sequences, with Applications to Audio-to-{MIDI} Alignment and Matching},
  year=2016
}"""


class LakhMIDIDataset(RemoteFolderDataset):
    """Lakh MIDI Dataset."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE, _LICENSE)
    _citation = _CITATION
    _sources = {
        "lmd_full": {
            "filename": "lmd_full.tar.gz",
            "url": "http://hog.ee.columbia.edu/craffel/lmd/lmd_full.tar.gz",
            "archive": True,
            "size": 1768163879,
            "md5": "2536ce3fd2cede53ddaa264f731859ab",
            "sha256": "6fcfe2ac49ca08f3f214cec86ab138d4fc4dabcd7f27f491a838dae6db45a12b",
        }
    }
    _extension = "mid"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)


class LakhMIDIMatchedDataset(RemoteFolderDataset):
    """Lakh MIDI Dataset - matched subset."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE, _LICENSE)
    _citation = _CITATION
    _sources = {
        "lmd_matched": {
            "filename": "lmd_matched.tar.gz",
            "url": "http://hog.ee.columbia.edu/craffel/lmd/lmd_matched.tar.gz",
            "archive": True,
            "size": 1407072670,
            "md5": "fb80d01c22020295bb3eeef31f1aa63a",
            "sha256": "621ff830aed771f469e5bfa13dc12a33c6ed69090adeda63d0b5c47783af0191",
        }
    }
    _extension = "mid"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)


class LakhMIDIAlignedDataset(RemoteFolderDataset):
    """Lakh MIDI Dataset - aligned subset."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE, _LICENSE)
    _citation = _CITATION
    _sources = {
        "lmd_aligned": {
            "filename": "lmd_aligned.tar.gz",
            "url": "http://hog.ee.columbia.edu/craffel/lmd/lmd_aligned.tar.gz",
            "archive": True,
            "size": 272169548,
            "md5": "d36ca9159966d81d97e1e37d10ed4584",
            "sha256": "2bf5400e82eba73204644946515489b68811e1e656b0cfd854efc14377f6e53b",
        }
    }
    _extension = "mid"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)



"""Piano-roll output interface."""
from operator import attrgetter
from typing import TYPE_CHECKING, Union

import numpy as np
from numpy import ndarray
from pypianoroll import Multitrack, Track

from ..classes import DEFAULT_VELOCITY

if TYPE_CHECKING:
    from ..music import Music


def to_pypianoroll(music: "Music") -> Multitrack:
    """Return a Music object as a Multitrack object.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to convert.

    Returns
    -------
    multitrack : :class:`pypianoroll.Multitrack`
        Converted Multitrack object.

    """
    length = (music.get_end_time() // music.resolution + 1) * music.resolution

    # Tracks
    tracks = []
    for track in music.tracks:
        pianoroll = np.zeros((length, 128))
        for note in track.notes:
            if note.velocity is not None:
                pianoroll[note.time : note.end, note.pitch] = note.velocity
            else:
                pianoroll[note.time : note.end, note.pitch] = DEFAULT_VELOCITY
        track = Track(
            program=track.program,
            is_drum=track.is_drum,
            name=track.name if track.name is not None else "",
            pianoroll=pianoroll,
        )
        tracks.append(track)

    # Tempos
    if not music.tempos:
        tempo_arr = None
    else:
        tempo_arr = np.full(length, 120.0)
        qpm = 120.0
        position = 0
        for tempo in music.tempos:
            tempo_arr[position : tempo.time] = qpm
            tempo_arr[tempo.time] = tempo.qpm
            position = tempo.time + 1
            qpm = tempo.qpm
        tempo_arr[position:] = qpm

    # Downbeats
    if not music.beats:
        downbeat_arr = None
    else:
        downbeat_arr = np.ones(length, bool)
        for beat in music.beats:
            if beat.is_downbeat:
                downbeat_arr[beat.time] = True

    has_title = music.metadata is not None and music.metadata.title is not None

    return Multitrack(
        name=music.metadata.title if has_title else None,
        resolution=music.resolution,
        tempo=tempo_arr,
        downbeat=downbeat_arr,
        tracks=tracks,
    )


def to_pianoroll_representation(
    music: "Music",
    encode_velocity: bool = True,
    dtype: Union[np.dtype, type, str] = None,
) -> ndarray:
    """Encode notes into piano-roll representation.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to encode.
    encode_velocity : bool, default: True
        Whether to encode velocities. If True, a binary-valued array
        will be return. Otherwise, an integer array will be return.
    dtype : np.dtype, type or str, optional
        Data type of the return array. Defaults to uint8 if
        `encode_velocity` is True, otherwise bool.

    Returns
    -------
    ndarray, shape=(?, 128)
        Encoded array in piano-roll representation.

    """
    if dtype is None:
        dtype = np.uint8 if encode_velocity else bool

    # Collect notes
    notes = []
    for track in music.tracks:
        notes.extend(track.notes)

    # Raise an error if no notes are found
    if not notes:
        raise RuntimeError("No notes found.")

    # Sort the notes
    notes.sort(key=attrgetter("time", "pitch", "duration", "velocity"))

    if not notes:
        return np.zeros((0, 128), dtype)

    # Initialize the array
    length = max((note.end for note in notes))
    array = np.zeros((length + 1, 128), dtype)

    # Encode notes
    for note in notes:
        if note.velocity is not None:
            if encode_velocity:
                array[note.time : note.end, note.pitch] = note.velocity
            else:
                array[note.time : note.end, note.pitch] = note.velocity > 0
        elif encode_velocity:
            array[note.time : note.end, note.pitch] = DEFAULT_VELOCITY
        else:
            array[note.time : note.end, note.pitch] = True

    return array



"""Haydn Op.20 Dataset."""
from pathlib import Path
from typing import List, Union

import music21
from music21.roman import RomanNumeral

from ..classes import Annotation
from ..inputs import from_music21_score
from ..music import DEFAULT_RESOLUTION, Music
from .base import DatasetInfo, RemoteFolderDataset

# pylint: disable=line-too-long

_NAME = "Haydn Op.20 Dataset"
_DESCRIPTION = """\
This dataset is a set of functional harmonic analysis annotations \
for the Op.20 string quartets from Joseph Haydn, commonly known as \
the 'Sun' quartets."""
_HOMEPAGE = "https://doi.org/10.5281/zenodo.1095630"
_CITATION = """\
@dataset{nestor_napoles_lopez_2017_1095630,
  author={N\'apoles L\'opez, N\'estor},
  title={{Joseph Haydn - String Quartets Op.20 - Harmonic Analysis Annotations Dataset}},
  month=dec,
  year=2017,
  publisher={Zenodo},
  version={v1.1-alpha},
  doi={10.5281/zenodo.1095630},
  url={https://doi.org/10.5281/zenodo.1095630}
}"""


class HaydnOp20Dataset(RemoteFolderDataset):
    """Haydn Op.20 Dataset."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _citation = _CITATION
    _sources = {
        "haydn": {
            "filename": "haydnop20v1.3_annotated.zip",
            "url": "https://github.com/napulen/haydn_op20_harm/releases/download/v1.3/haydnop20v1.3_annotated.zip",
            "archive": True,
            "size": 130954,
            "md5": "1c65c8da312e1c9dda681d0496bf527f",
            "sha256": "96986cccebfd37a36cc97a2fc0ebcfbe22d5136e622b21e04ea125d589f5073b",
        }
    }
    _extension = "hrm"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        score = music21.converter.parse(filename, format="humdrum")

        # Get the annotations
        roman_numerals = list(score.flat.getElementsByClass("RomanNumeral"))
        annotations = get_annotations(roman_numerals)

        # Remove the annotations from the original score
        # (they mess with the python representation)
        score.remove(roman_numerals, recurse=True)

        music = from_music21_score(score)
        music.annotations = annotations

        return music


def get_annotations(
    roman_numerals: List[RomanNumeral], resolution=DEFAULT_RESOLUTION
) -> List[Annotation]:
    """Return music21 RomanNumeral objects as Annotation objects."""
    # Convert the list into a dictionary to remove duplicate items
    roman_numeral_dict = {rn.offset: rn for rn in roman_numerals if rn}

    annotations = []
    for offset, roman_numeral in roman_numeral_dict.items():
        time = int(round(float(offset * resolution)))
        tonicized_key = roman_numeral.secondaryRomanNumeralKey
        key = tonicized_key if tonicized_key else roman_numeral.key

        annotation = {
            "key": key.tonicPitchNameWithCase,
            "figure": roman_numeral.figure,
            "chord": roman_numeral.pitchedCommonName,
        }
        annotations.append(Annotation(time=time, annotation=annotation))

    return annotations



"""Note-based representation output interface."""
from operator import attrgetter
from typing import TYPE_CHECKING, Union

import numpy as np
from numpy import ndarray

from ..classes import DEFAULT_VELOCITY

if TYPE_CHECKING:
    from ..music import Music


def to_note_representation(
    music: "Music",
    use_start_end: bool = False,
    encode_velocity: bool = True,
    dtype: Union[np.dtype, type, str] = int,
) -> ndarray:
    """Encode a Music object into note-based representation.

    The note-based represetantion represents music as a sequence of
    (time, pitch, duration, velocity) tuples. For example, a note
    Note(time=0, duration=4, pitch=60, velocity=64) will be encoded as a
    tuple (0, 60, 4, 64). The output shape is N * D, where N is the
    number of notes and D is 4 when `encode_velocity` is True, otherwise
    D is 3. The values of the second dimension represent time, pitch,
    duration and velocity (discarded when `encode_velocity` is False).

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to encode.
    use_start_end : bool, default: False
        Whether to use 'start' and 'end' to encode the timing rather
        than 'time' and 'duration'.
    encode_velocity : bool, default: True
        Whether to encode note velocities.
    dtype : np.dtype, type or str, default: int
        Data type of the return array.

    Returns
    -------
    ndarray, shape=(?, 3 or 4)
        Encoded array in note-based representation.

    """
    # Collect notes
    notes = []
    for track in music.tracks:
        notes.extend(track.notes)

    # Raise an error if no notes is found
    if not notes:
        raise RuntimeError("No notes found.")

    # Sort the notes
    notes.sort(key=attrgetter("time", "pitch", "duration", "velocity"))

    # Initialize the array
    if encode_velocity:
        array = np.zeros((len(notes), 4), dtype)
    else:
        array = np.zeros((len(notes), 3), dtype)

    # Encode notes
    for i, note in enumerate(notes):
        array[i, 0] = note.time
        array[i, 1] = note.pitch
        array[i, 2] = note.end if use_start_end else note.duration
        if encode_velocity:
            if note.velocity is not None:
                array[i, 3] = note.velocity
            else:
                array[i, 3] = DEFAULT_VELOCITY

    return array



"""Wrapper functions for input interface."""
from pathlib import Path
from typing import List, TextIO, Union

from mido import MidiFile
from music21.stream import Stream
from numpy import ndarray
from pretty_midi import PrettyMIDI
from pypianoroll import Multitrack

from ..classes import Track
from ..music import Music
from .abc import read_abc
from .event import from_event_representation
from .json import load_json
from .midi import from_mido, from_pretty_midi, read_midi
from .music21 import from_music21
from .musicxml import read_musicxml
from .note import from_note_representation
from .pianoroll import from_pianoroll_representation, from_pypianoroll
from .pitch import from_pitch_representation
from .yaml import load_yaml


def load(path: Union[str, Path, TextIO], kind: str = None, **kwargs) -> Music:
    """Load a JSON or a YAML file into a Music object.

    This is a wrapper function for :func:`muspy.load_json` and
    :func:`muspy.load_yaml`.

    Parameters
    ----------
    path : str, Path or TextIO
        Path to the file or the file to to load.
    kind : {'json', 'yaml'}, optional
        Format to save. Defaults to infer from the extension.
    **kwargs
        Keyword arguments to pass to :func:`muspy.load_json` or
        :func:`muspy.load_yaml`.

    Returns
    -------
    :class:`muspy.Music`
        Loaded Music object.

    See Also
    --------
    :func:`muspy.load_json` : Load a JSON file into a Music object.
    :func:`muspy.load_yaml` : Load a YAML file into a Music object.
    :func:`muspy.read` :
        Read a MIDI/MusicXML/ABC file into a Music object.

    """
    if kind is None:
        if not isinstance(path, (str, Path)):
            raise ValueError("Cannot infer file format from a file object.")
        path_str = str(path).lower()
        if path_str.endswith((".json", ".json.gz")):
            kind = "json"
        elif path_str.endswith((".yaml", ".yml", ".yaml.gz", ".yml.gz")):
            kind = "yaml"
        else:
            raise ValueError(
                "Cannot infer file format from the extension (expect JSON or "
                "YAML)."
            )
    if kind.lower() == "json":
        return load_json(path, **kwargs)
    if kind.lower() == "yaml":
        return load_yaml(path, **kwargs)
    raise ValueError(
        f"Expect `kind` to be 'json' or 'yaml', but got : {kind}."
    )


def read(
    path: Union[str, Path], kind: str = None, **kwargs
) -> Union[Music, List[Music]]:
    """Read a MIDI/MusicXML/ABC file into a Music object.

    Parameters
    ----------
    path : str or Path
        Path to the file to read.
    kind : {'midi', 'musicxml', 'abc'}, optional
        Format to save. Defaults to infer from the extension.
    **kwargs
        Keyword arguments to pass to :func:`muspy.read_midi`,
        :func:`muspy.read_musicxml` or :func:`read_abc`.

    Returns
    -------
    :class:`muspy.Music` or list of :class:`muspy.Music`
        Converted Music object(s).

    See Also
    --------
    :func:`muspy.load` : Load a JSON or a YAML file into a Music object.

    """
    if kind is None:
        if str(path).lower().endswith((".mid", ".midi")):
            kind = "midi"
        elif str(path).lower().endswith((".mxl", ".xml", ".musicxml")):
            kind = "musicxml"
        elif str(path).lower().endswith(".abc"):
            kind = "abc"
        else:
            raise ValueError(
                "Cannot infer file format from the extension (expect MIDI, "
                "MusicXML or ABC)."
            )
    if kind.lower() == "midi":
        return read_midi(path, **kwargs)
    if kind.lower() == "musicxml":
        return read_musicxml(path, **kwargs)
    if kind.lower() == "abc":
        return read_abc(path, **kwargs)
    raise ValueError(
        f"Expect `kind` to be 'midi', 'musicxml' or 'abc', but got : {kind}."
    )


def from_object(
    obj: Union[Stream, MidiFile, PrettyMIDI, Multitrack], **kwargs
) -> Union[Music, List[Music], Track, List[Track]]:
    """Return an outside object as a Music object.

    Parameters
    ----------
    obj
        Object to convert. Supported objects are `music21.Stream`,
        :class:`mido.MidiTrack`, :class:`pretty_midi.PrettyMIDI`, and
        :class:`pypianoroll.Multitrack` objects.
    **kwargs
        Keyword arguments to pass to :func:`muspy.from_music21`,
        :func:`muspy.from_mido`, :func:`from_pretty_midi` or
        :func:`from_pypianoroll`.

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    if isinstance(obj, Stream):
        return from_music21(obj, **kwargs)
    if isinstance(obj, MidiFile):
        return from_mido(obj, **kwargs)
    if isinstance(obj, PrettyMIDI):
        return from_pretty_midi(obj, **kwargs)
    if isinstance(obj, Multitrack):
        return from_pypianoroll(obj, **kwargs)
    raise TypeError(
        "`obj` must be of type music21.Stream, mido.MidiFile, "
        "pretty_midi.PrettyMIDI or pypianoroll.Multitrack."
    )


def from_representation(array: ndarray, kind: str, **kwargs) -> Music:
    """Update with the given representation.

    Parameters
    ----------
    array : :class:`numpy.ndarray`
        Array in a supported representation.
    kind : str, {'pitch', 'pianoroll', 'event', 'note'}
        Data representation.
    **kwargs
        Keyword arguments to pass to
        :func:`muspy.from_pitch_representation`,
        :func:`muspy.from_pianoroll_representation`,
        :func:`from_event_representation` or
        :func:`from_note_representation`.

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    if kind.lower() in ("pitch", "pitch-based"):
        return from_pitch_representation(array, **kwargs)
    if kind.lower() in ("pianoroll", "piano-roll", "piano roll"):
        return from_pianoroll_representation(array, **kwargs)
    if kind.lower() in ("event", "event-based"):
        return from_event_representation(array, **kwargs)
    if kind.lower() in ("note", "note-based"):
        return from_note_representation(array, **kwargs)
    raise ValueError(
        "Expect `kind` to be 'pitch', 'pianoroll', 'event' or 'note', but"
        f"got : {kind}."
    )



"""Base classes.

This module defines the bases classes for MusPy objects.

Classes
-------

- Base
- ComplexBase

"""
import copy
from collections import OrderedDict
from inspect import isclass
from typing import Any, Callable, Iterable, List, Mapping, Type, TypeVar, Union

from .utils import yaml_dump

__all__ = ["Base", "ComplexBase"]

BaseType = TypeVar("BaseType", bound="Base")
ComplexBaseType = TypeVar("ComplexBaseType", bound="ComplexBase")


def _get_type_string(attr_type):
    """Return a string represeting acceptable type(s)."""
    if isinstance(attr_type, (list, tuple)):
        if len(attr_type) > 1:
            return (
                ", ".join([x.__name__ for x in attr_type[:-1]])
                + " or "
                + attr_type[-1].__name__
            )
        return attr_type[0].__name__
    return attr_type.__name__


class Base:
    """Base class for MusPy classes.

    This is the base class for MusPy classes. It provides two handy I/O
    methods---`from_dict` and `to_ordered_dict`. It also provides
    intuitive `repr` as well as methods `pretty_str` and `print` for
    beautifully printing the content.

    In addition, `hash` is implemented by `hash(repr(self))`.
    Comparisons between two Base objects are also supported, where
    equality check will compare all attributes, while 'less than' and
    'greater than' will only compare the `time` attribute.

    Hint
    ----
    To implement a new class in MusPy, please inherit from this class
    and set the following class variables properly.

    - `_attributes`: An OrderedDict with attribute names as keys and
      their types as values.
    - `_optional_attributes`: A list of optional attribute names.
    - `_list_attributes`: A list of attributes that are lists.

    Take :class:`muspy.Note` for example.::

        _attributes = OrderedDict(
            [
                ("time", int),
                ("duration", int),
                ("pitch", int),
                ("velocity", int),
                ("pitch_str", str),
            ]
        )
        _optional_attributes = ["pitch_str"]

    See Also
    --------
    :class:`muspy.ComplexBase` :
        Base class that supports advanced operations on list attributes.

    """

    _attributes: Mapping[str, Any] = {}
    _optional_attributes: List[str] = []
    _list_attributes: List[str] = []

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        to_join = []
        for attr in self._attributes:
            value = getattr(self, attr)
            if attr in self._list_attributes:
                if not value:
                    continue
                if len(value) > 3:
                    to_join.append(
                        attr + "=" + repr(value[:3])[:-1] + ", ...]"
                    )
                else:
                    to_join.append(attr + "=" + repr(value))
            elif value is not None:
                to_join.append(attr + "=" + repr(value))
        return type(self).__name__ + "(" + ", ".join(to_join) + ")"

    def __hash__(self) -> int:
        return hash(repr(self))

    def __eq__(self, other) -> bool:
        for attr in self._attributes:
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True

    def __lt__(self, other) -> bool:
        if not hasattr(self, "time") or not hasattr(other, "time"):
            return NotImplemented
        if getattr(self, "time") < getattr(other, "time"):
            return True
        return False

    def __gt__(self, other) -> bool:
        if not hasattr(self, "time") or not hasattr(other, "time"):
            return NotImplemented
        if getattr(self, "time") > getattr(other, "time"):
            return True
        return False

    def __deepcopy__(self: BaseType, memo: dict) -> BaseType:
        return self.from_dict(self.to_ordered_dict())

    @classmethod
    def from_dict(
        cls: Type[BaseType],
        dict_: Mapping,
        strict: bool = False,
        cast: bool = False,
    ) -> BaseType:
        """Return an instance constructed from a dictionary.

        Instantiate an object whose attributes and the corresponding
        values are given as a dictionary.

        Parameters
        ----------
        dict_ : dict or mapping
            A dictionary that stores the attributes and their values as
            key-value pairs, e.g., `{"attr1": value1, "attr2": value2}`.
        strict : bool, default: False
            Whether to raise errors for invalid input types.
        cast : bool, default: False
            Whether to cast types.

        Returns
        -------
        Constructed object.

        """
        assert not (
            strict and cast
        ), "`strict` and `cast` cannot be both True."
        kwargs = {}
        for attr, attr_type in cls._attributes.items():
            if isinstance(attr_type, tuple):
                attr_type = attr_type[0]
            value = dict_.get(attr)
            if value is None:
                if attr in cls._optional_attributes:
                    continue
                raise TypeError(f"`{attr}` must not be None.")
            if isclass(attr_type) and issubclass(attr_type, Base):
                if attr in cls._list_attributes:
                    kwargs[attr] = [attr_type.from_dict(v) for v in value]
                else:
                    kwargs[attr] = attr_type.from_dict(value)
            else:
                if strict:
                    if attr in cls._list_attributes:
                        if not isinstance(value, list):
                            raise TypeError(
                                f"`{attr}` must be a list, but got : "
                                f"{type(value)} ."
                            )
                        for v in value:  # pylint: disable=invalid-name
                            if not isinstance(v, attr_type):
                                raise TypeError(
                                    f"`{attr}` must be a list of type "
                                    f"{attr_type}, but got : {type(v)} ."
                                )
                    elif not isinstance(value, attr_type):
                        raise TypeError(
                            f"`{attr}` must be of type {attr_type}, "
                            f"but got : {type(value)} ."
                        )
                if cast:
                    is_bad_input = False
                    if attr in cls._list_attributes:
                        if not isinstance(value, list):
                            is_bad_input = True
                        for v in value:  # pylint: disable=invalid-name
                            if not isinstance(v, attr_type):
                                is_bad_input = True
                        if is_bad_input:
                            kwargs[attr] = [attr_type(v) for v in value]
                        else:
                            kwargs[attr] = value
                    elif not isinstance(value, attr_type):
                        kwargs[attr] = attr_type(value)
                else:
                    kwargs[attr] = value
        return cls(**kwargs)

    def to_ordered_dict(
        self, skip_missing: bool = True, deepcopy: bool = True,
    ) -> OrderedDict:
        """Return the object as an OrderedDict.

        Return an ordered dictionary that stores the attributes and
        their values as key-value pairs.

        Parameters
        ----------
        skip_missing : bool, default: True
            Whether to skip attributes with value None or those that are
            empty lists.
        deepcopy : bool, default: True
            Whether to make deep copies of the attributes.

        Returns
        -------
        OrderedDict
            A dictionary that stores the attributes and their values as
            key-value pairs, e.g., `{"attr1": value1, "attr2": value2}`.

        """
        ordered_dict: OrderedDict = OrderedDict()
        for attr, attr_type in self._attributes.items():
            value = getattr(self, attr)
            if attr in self._list_attributes:
                if not value and skip_missing:
                    continue
                if isclass(attr_type) and issubclass(attr_type, Base):
                    ordered_dict[attr] = [
                        v.to_ordered_dict(
                            skip_missing=skip_missing, deepcopy=deepcopy
                        )
                        for v in value
                    ]
                elif deepcopy:
                    ordered_dict[attr] = copy.deepcopy(value)
                else:
                    ordered_dict[attr] = value
            elif value is None:
                if not skip_missing:
                    ordered_dict[attr] = None
            elif isclass(attr_type) and issubclass(attr_type, Base):
                ordered_dict[attr] = value.to_ordered_dict(
                    skip_missing=skip_missing, deepcopy=deepcopy
                )
            elif deepcopy:
                ordered_dict[attr] = copy.deepcopy(value)
            else:
                ordered_dict[attr] = value
        return ordered_dict

    def copy(self: BaseType) -> BaseType:
        """Return a shallow copy of the object.

        This is equivalent to :py:func:`copy.copy(self)`.

        Returns
        -------
        Shallow copy of the object.

        """
        return copy.copy(self)

    def deepcopy(self: BaseType) -> BaseType:
        """Return a deep copy of the object.

        This is equivalent to :py:func:`copy.deepcopy(self)`

        Returns
        -------
        Deep copy of the object.

        """
        return copy.deepcopy(self)

    def pretty_str(self, skip_missing: bool = True) -> str:
        """Return the attributes as a string in a YAML-like format.

        Parameters
        ----------
        skip_missing : bool, default: True
            Whether to skip attributes with value None or those that are
            empty lists.

        Returns
        -------
        str
            Stored data as a string in a YAML-like format.

        See Also
        --------
        :meth:`muspy.Base.print` :
            Print the attributes in a YAML-like format.

        """
        return yaml_dump(
            self.to_ordered_dict(skip_missing=skip_missing, deepcopy=False)
        )

    def print(self, skip_missing: bool = True):
        """Print the attributes in a YAML-like format.

        Parameters
        ----------
        skip_missing : bool, default: True
            Whether to skip attributes with value None or those that are
            empty lists.

        See Also
        --------
        :meth:`muspy.Base.pretty_str` :
            Return the the attributes as a string in a YAML-like format.

        """
        print(self.pretty_str(skip_missing=skip_missing))

    def _validate_attr_type(self, attr: str, recursive: bool):
        attr_type = self._attributes[attr]
        value = getattr(self, attr)
        if value is None:
            if attr in self._optional_attributes:
                return
            raise TypeError("`{}` must not be None".format(attr))
        if attr in self._list_attributes:
            if not isinstance(value, list):
                raise TypeError("`{}` must be a list.".format(attr))
            for item in value:
                if not isinstance(item, attr_type):
                    raise TypeError(
                        "`{}` must be a list of type {}.".format(
                            attr, _get_type_string(attr_type)
                        )
                    )
        elif not isinstance(value, attr_type):
            raise TypeError(
                "`{}` must be of type {}.".format(
                    attr, _get_type_string(attr_type)
                )
            )

        # Apply recursively
        if recursive and isclass(attr_type) and issubclass(attr_type, Base):
            if attr in self._list_attributes:
                for item in getattr(self, attr):
                    item.validate_type(recursive=recursive)
            elif getattr(self, attr) is not None:
                getattr(self, attr).validate_type(recursive=recursive)

    def validate_type(
        self: BaseType, attr: str = None, recursive: bool = True,
    ) -> BaseType:
        """Raise an error if an attribute is of an invalid type.

        This will apply recursively to an attribute's attributes.

        Parameters
        ----------
        attr : str, optional
            Attribute to validate. Defaults to validate all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        See Also
        --------
        :meth:`muspy.Base.is_valid_type` :
            Return True if an attribute is of a valid type.
        :meth:`muspy.Base.validate` :
            Raise an error if an attribute has an invalid type or value.

        """
        if attr is None:
            for attribute in self._attributes:
                self._validate_attr_type(attribute, recursive)
        else:
            self._validate_attr_type(attr, recursive)
        return self

    def _validate(self, attr: str, recursive: bool):
        attr_type = self._attributes[attr]
        if isclass(attr_type) and issubclass(attr_type, Base):
            if attr in self._list_attributes:
                if getattr(self, attr):
                    for item in getattr(self, attr):
                        item.validate()
            else:
                getattr(self, attr).validate()
        else:
            # Set recursive=False to avoid repeated checks invoked when
            # calling `validate` recursively
            self._validate_attr_type(attr, False)
            if attr == "time" and getattr(self, "time") < 0:
                raise ValueError("`time` must be nonnegative.")

        # Apply recursively
        if recursive and isclass(attr_type) and issubclass(attr_type, Base):
            if attr in self._list_attributes:
                for item in getattr(self, attr):
                    item.validate(recursive=recursive)
            elif getattr(self, attr) is not None:
                getattr(self, attr).validate(recursive=recursive)

    def validate(
        self: BaseType, attr: str = None, recursive: bool = True,
    ) -> BaseType:
        """Raise an error if an attribute has an invalid type or value.

        This will apply recursively to an attribute's attributes.

        Parameters
        ----------
        attr : str, optional
            Attribute to validate. Defaults to validate all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        See Also
        --------
        :meth:`muspy.Base.is_valid` :
            Return True if an attribute has a valid type and value.
        :meth:`muspy.Base.validate_type` :
            Raise an error if an attribute is of an invalid type.

        """
        if attr is None:
            for attribute in self._attributes:
                self._validate(attribute, recursive)
        else:
            self._validate(attr, recursive)
        return self

    def is_valid_type(self, attr: str = None, recursive: bool = True,) -> bool:
        """Return True if an attribute is of a valid type.

        This will apply recursively to an attribute's attributes.

        Parameters
        ----------
        attr : str, optional
            Attribute to validate. Defaults to validate all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        bool
            Whether the attribute is of a valid type.

        See Also
        --------
        :meth:`muspy.Base.validate_type` :
            Raise an error if a certain attribute is of an invalid type.
        :meth:`muspy.Base.is_valid` :
            Return True if an attribute has a valid type and value.

        """
        try:
            self.validate_type(attr, recursive)
        except TypeError:
            return False
        return True

    def is_valid(self, attr: str = None, recursive: bool = True,) -> bool:
        """Return True if an attribute has a valid type and value.

        This will recursively apply to an attribute's attributes.

        Parameters
        ----------
        attr : str, optional
            Attribute to validate. Defaults to validate all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        bool
            Whether the attribute has a valid type and value.

        See Also
        --------
        :meth:`muspy.Base.validate` :
            Raise an error if an attribute has an invalid type or value.
        :meth:`muspy.Base.is_valid_type` :
            Return True if an attribute is of a valid type.

        """
        try:
            self.validate(attr, recursive)
        except (TypeError, ValueError):
            return False
        return True

    def _adjust_time(
        self, func: Callable[[int], int], attr: str, recursive: bool
    ):
        attr_type = self._attributes[attr]
        if attr == "time":
            if "time" in self._list_attributes:
                new_list = [func(item) for item in getattr(self, "time")]
                setattr(self, "time", new_list)
            else:
                setattr(self, "time", func(getattr(self, attr)))
        elif recursive and isclass(attr_type) and issubclass(attr_type, Base):
            if attr in self._list_attributes:
                for item in getattr(self, attr):
                    item.adjust_time(func, recursive=recursive)
            elif getattr(self, attr) is not None:
                getattr(self, attr).adjust_time(func, recursive=recursive)

    def adjust_time(
        self: BaseType,
        func: Callable[[int], int],
        attr: str = None,
        recursive: bool = True,
    ) -> BaseType:
        """Adjust the timing of time-stamped objects.

        Parameters
        ----------
        func : callable
            The function used to compute the new timing from the old
            timing, i.e., `new_time = func(old_time)`.
        attr : str, optional
            Attribute to adjust. Defaults to adjust all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        """
        if attr is None:
            for attribute in self._attributes:
                self._adjust_time(func, attribute, recursive)
        else:
            self._adjust_time(func, attr, recursive)
        return self

    def _fix_type(self: BaseType, attr: str, recursive: bool):
        attr_type = self._attributes[attr]
        if isclass(attr_type) and issubclass(attr_type, Base):
            if attr in self._list_attributes:
                if getattr(self, attr):
                    for item in getattr(self, attr):
                        item.fix_type()
            else:
                getattr(self, attr).fix_type()
        else:
            value = getattr(self, attr)
            if not isinstance(value, attr_type):
                if isinstance(attr_type, tuple):
                    setattr(self, attr, attr_type[0](value))
                else:
                    setattr(self, attr, attr_type(value))

        # Apply recursively
        if recursive and isclass(attr_type) and issubclass(attr_type, Base):
            if attr in self._list_attributes:
                for item in getattr(self, attr):
                    item.fix_type(recursive=recursive)
            elif getattr(self, attr) is not None:
                getattr(self, attr).fix_type(recursive=recursive)

    def fix_type(
        self: BaseType, attr: str = None, recursive: bool = True
    ) -> BaseType:
        """Fix the types of attributes.

        Parameters
        ----------
        attr : str, optional
            Attribute to adjust. Defaults to adjust all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        """
        if attr is None:
            for attribute in self._attributes:
                self._fix_type(attribute, recursive)
        else:
            self._fix_type(attr, recursive)
        return self


class ComplexBase(Base):
    """Base class that supports advanced operations on list attributes.

    This class extend the Base class with advanced operations on list
    attributes, including `append`, `remove_invalid`, `remove_duplicate`
    and `sort`.

    See Also
    --------
    :class:`muspy.Base` : Base class for MusPy classes.

    """

    def __iadd__(
        self: ComplexBaseType, other: Union[ComplexBaseType, Iterable]
    ) -> ComplexBaseType:
        return self.extend(other)

    def __add__(
        self: ComplexBaseType, other: ComplexBaseType
    ) -> ComplexBaseType:
        if not isinstance(other, type(self)):
            raise TypeError(
                "Expect the second operand to be of type "
                f"{type(self).__name__}, but got {type(other).__name__}."
            )
        return self.deepcopy().extend(other, deepcopy=True)

    def _append(self, obj):
        for attr in self._list_attributes:
            attr_type = self._attributes[attr]
            if isinstance(obj, attr_type):
                if isclass(attr_type) and issubclass(attr_type, Base):
                    if getattr(self, attr) is None:
                        setattr(self, attr, [obj])
                    else:
                        getattr(self, attr).append(obj)
                    return
        raise TypeError(
            f"Cannot find a list attribute for type {type(obj).__name__}."
        )

    def append(self: ComplexBaseType, obj) -> ComplexBaseType:
        """Append an object to the corresponding list.

        This will automatically determine the list attributes to append
        based on the type of the object.

        Parameters
        ----------
        obj
            Object to append.

        """
        self._append(obj)
        return self

    def extend(
        self: ComplexBaseType,
        other: Union[ComplexBaseType, Iterable],
        deepcopy: bool = False,
    ) -> ComplexBaseType:
        """Extend the list(s) with another object or iterable.

        Parameters
        ----------
        other : :class:`muspy.ComplexBase` or iterable
            If an object of the same type is given, extend the
            list attributes with the corresponding list attributes of
            the other object. If an iterable is given, call
            :meth:`muspy.ComplexBase.append` for each item.
        deepcopy : bool, default: False
            Whether to make deep copies of the appended objects.

        Returns
        -------
        Object itself.

        """
        if isinstance(other, ComplexBase):
            if not isinstance(other, type(self)):
                raise TypeError(
                    f"Expect `other` to be of type {type(self).__name__}, "
                    f"but got {type(other).__name__}."
                )
            for attr in self._list_attributes:
                other_value = getattr(other, attr)
                getattr(self, attr).extend(
                    copy.deepcopy(other_value) if deepcopy else other_value
                )
            return self

        for item in other:  # type: ignore
            self._append(copy.deepcopy(item) if deepcopy else item)
        return self

    def _remove_invalid(self, attr: str, recursive: bool):
        # Skip it if empty
        if not getattr(self, attr):
            return

        attr_type = self._attributes[attr]
        is_class = isclass(attr_type)
        is_base = is_class and issubclass(attr_type, Base)
        is_complexbase = is_class and issubclass(attr_type, ComplexBase)
        value = getattr(self, attr)

        # NOTE: The ordering mathers here. We first apply recursively
        # and later check the currect object so that something that can
        # be fixed in a lower level would not make the high-level object
        # removed.

        # Apply recursively
        if recursive and is_complexbase:
            for item in value:
                item.remove_invalid(recursive=recursive)

        # Replace the old list with a new list of only valid items
        if is_base:
            value[:] = [item for item in value if item.is_valid()]
        else:
            value[:] = [item for item in value if isinstance(item, attr_type)]

    def remove_invalid(
        self: ComplexBaseType, attr: str = None, recursive: bool = True,
    ) -> ComplexBaseType:
        """Remove invalid items from a list attribute.

        Parameters
        ----------
        attr : str, optional
            Attribute to validate. Defaults to validate all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        """
        if attr is None:
            for attribute in self._list_attributes:
                self._remove_invalid(attribute, recursive)
        elif attr in self._list_attributes:
            self._remove_invalid(attr, recursive)
        else:
            raise TypeError("`{}` must be a list attribute.")
        return self

    def _remove_duplicate(self, attr: str, recursive: bool):
        # Skip it if empty
        if not getattr(self, attr):
            return

        attr_type = self._attributes[attr]
        is_complexbase = isclass(attr_type) and issubclass(
            attr_type, ComplexBase
        )
        value = getattr(self, attr)

        # NOTE: The ordering mathers here. We first apply recursively
        # and later check the currect object so that something that can
        # be fixed in a lower level would not make the high-level object
        # removed.

        # Apply recursively
        if recursive and is_complexbase:
            for item in value:
                item.remove_duplicate(recursive=recursive)

        # Replace the old list with a new list without duplicates
        # TODO: Speed this up by grouping by time.
        new_value = []
        for item in value:
            if item not in new_value:
                new_value.append(item)
        value[:] = new_value

    def remove_duplicate(
        self: ComplexBaseType, attr: str = None, recursive: bool = True,
    ) -> ComplexBaseType:
        """Remove duplicate items from a list attribute.

        Parameters
        ----------
        attr : str, optional
            Attribute to check. Defaults to check all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        """
        if attr is None:
            for attribute in self._list_attributes:
                self._remove_duplicate(attribute, recursive)
        elif attr in self._list_attributes:
            self._remove_duplicate(attr, recursive)
        else:
            raise TypeError("`{}` must be a list attribute.")
        return self

    def _sort(self, attr: str, recursive: bool):
        # Skip it if empty
        if not getattr(self, attr):
            return

        # Sort the list
        attr_type = self._attributes[attr]
        if isclass(attr_type) and issubclass(attr_type, Base):
            getattr(self, attr).sort()
            # Apply recursively
            if recursive and issubclass(attr_type, ComplexBase):
                for value in getattr(self, attr):
                    value.sort(recursive=recursive)

    def sort(
        self: ComplexBaseType, attr: str = None, recursive: bool = True,
    ) -> ComplexBaseType:
        """Sort a list attribute.

        Parameters
        ----------
        attr : str, optional
            Attribute to sort. Defaults to sort all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        """
        if attr is None:
            for attribute in self._list_attributes:
                self._sort(attribute, recursive)
        elif attr in self._list_attributes:
            self._sort(attr, recursive)
        else:
            raise TypeError("`{}` must be a list attribute.")
        return self



"""Wikifonia Dataset."""
from pathlib import Path
from typing import Union

from ..inputs import read_musicxml
from ..music import Music
from .base import DatasetInfo, RemoteFolderDataset

# pylint: disable=line-too-long

_NAME = "Wikifonia Dataset"
_DESCRIPTION = """\
Wikifonia dataset is a collection of 6675 lead sheets in MusicMXL format. It
was originally hosted at http://www.wikifonia.org/."""
_HOMEPAGE = "http://www.wikifonia.org/"


class WikifoniaDataset(RemoteFolderDataset):
    """Wikifonia dataset."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _sources = {
        "wikifonia": {
            "filename": "Wikifonia.zip",
            "url": "http://www.synthzone.com/files/Wikifonia/Wikifonia.zip",
            "archive": True,
            "size": 35727800,
            "md5": "d26e22562e67eb7d37535e96cc5eebba",
            "sha256": "e7bce509462a73cee175308b6a3cdafa9effd6e8958b3ce03b4edb293cc6b691",
        }
    }
    _extension = "mxl"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_musicxml(filename)



"""MAESTRO Dataset."""
from pathlib import Path
from typing import Union

from ..inputs import read_midi
from ..music import Music
from .base import DatasetInfo, RemoteFolderDataset

# pylint: disable=line-too-long

_NAME = "MAESTRO Dataset"
_DESCRIPTION = """\
MAESTRO (MIDI and Audio Edited for Synchronous TRacks and Organization) is a \
dataset composed of over 200 hours of virtuosic piano performances captured \
with fine alignment (~3 ms) between note labels and audio waveforms."""
_HOMEPAGE = "https://magenta.tensorflow.org/datasets/maestro"
_LICENSE = "Creative Commons Attribution Non-Commercial Share-Alike 4.0 \
(CC BY-NC-SA 4.0)"
_CITATION = """\
@inproceedings{hawthorne2018enabling,
  title={Enabling Factorized Piano Music Modeling and Generation with the {MAESTRO} Dataset},
  author={Curtis Hawthorne and Andriy Stasyuk and Adam Roberts and Ian Simon and Cheng-Zhi Anna Huang and Sander Dieleman and Erich Elsen and Jesse Engel and Douglas Eck},
  booktitle={Proceedings of the 7th International Conference on Learning Representations (ICLR)},
  year=2019,
  url={https://openreview.net/forum?id=r1lYRjC9F7}
}"""


class MAESTRODatasetV1(RemoteFolderDataset):
    """MAESTRO Dataset V1 (MIDI only)."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE, _LICENSE)
    _citation = _CITATION
    _sources = {
        "maestro": {
            "filename": "maestro-v1.0.0-midi.zip",
            "url": "https://storage.googleapis.com/magentadata/datasets/maestro/v1.0.0/maestro-v1.0.0-midi.zip",
            "archive": True,
            "size": 46579421,
            "sha256": "f620f9e1eceaab8beea10617599add2e9c83234199b550382a2f603098ae7135",
        }
    }
    _extension = "midi"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)


class MAESTRODatasetV2(RemoteFolderDataset):
    """MAESTRO Dataset V2 (MIDI only)."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE, _LICENSE)
    _citation = _CITATION
    _sources = {
        "maestro": {
            "filename": "maestro-v2.0.0-midi.zip",
            "url": "https://storage.googleapis.com/magentadata/datasets/maestro/v2.0.0/maestro-v2.0.0-midi.zip",
            "archive": True,
            "size": 59243107,
            "sha256": "ec2cc9d94886c6b376db1eaa2b8ad1ce62ff9f0a28b3744782b13163295dadf3",
        }
    }
    _extension = "midi"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)


class MAESTRODatasetV3(RemoteFolderDataset):
    """MAESTRO Dataset V3 (MIDI only)."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE, _LICENSE)
    _citation = _CITATION
    _sources = {
        "maestro": {
            "filename": "maestro-v3.0.0-midi.zip",
            "url": "https://storage.googleapis.com/magentadata/datasets/maestro/v3.0.0/maestro-v3.0.0-midi.zip",
            "archive": True,
            "size": 58416533,
            "sha256": "70470ee253295c8d2c71e6d9d4a815189e35c89624b76d22fce5a019d5dde12c",
        }
    }
    _extension = "midi"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)



"""Pitch-based representation input interface."""
from operator import attrgetter
from typing import List

import numpy as np
from numpy import ndarray

from ..classes import DEFAULT_VELOCITY, Note, Track
from ..music import DEFAULT_RESOLUTION, Music


def from_pitch_representation(
    array: ndarray,
    resolution: int = DEFAULT_RESOLUTION,
    program: int = 0,
    is_drum: bool = False,
    use_hold_state: bool = False,
    default_velocity: int = DEFAULT_VELOCITY,
) -> Music:
    """Decode pitch-based representation into a Music object.

    Parameters
    ----------
    array : ndarray
        Array in pitch-based representation to decode.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.
    program : int, default: 0 (Acoustic Grand Piano)
        Program number, according to General MIDI specification [1].
        Valid values are 0 to 127.
    is_drum : bool, default: False
        Whether it is a percussion track.
    use_hold_state : bool, default: False
        Whether to use a special state for holds.
    default_velocity : int, default: `muspy.DEFAULT_VELOCITY` (64)
        Default velocity value to use when decoding.

    Returns
    -------
    :class:`muspy.Music`
        Decoded Music object.

    References
    ----------
    [1] https://www.midi.org/specifications/item/gm-level-1-sound-set

    """
    if not np.issubdtype(array.dtype, np.integer):
        raise TypeError("Array must be of type int.")

    # Find the note boundaries
    notes: List[Note] = []
    diff = np.diff(array.flatten(), prepend=-1, append=-1)
    boundaries = np.nonzero(diff)[0]

    # Decode pitches
    if use_hold_state:
        is_awaiting_hold = False
        for start, end in zip(boundaries[:-1], boundaries[1:]):
            # Skip rests
            if array[start] == 128:
                is_awaiting_hold = False
                continue

            # Hold
            if array[start] == 129:
                # Skip a hold that does not follow any pitch
                if not is_awaiting_hold:
                    continue

                notes[-1].duration += end - start
                is_awaiting_hold = False

            # Pitch
            else:
                note = Note(
                    time=int(start),
                    pitch=int(array[start]),
                    duration=int(end - start),
                    velocity=default_velocity,
                )
                notes.append(note)
                is_awaiting_hold = True

    else:
        for start, end in zip(boundaries[:-1], boundaries[1:]):
            # Skip rests
            if array[start] == 128:
                continue

            # Pitch
            note = Note(
                time=int(start),
                pitch=int(array[start]),
                duration=int(end - start),
                velocity=default_velocity,
            )
            notes.append(note)

    # Sort the notes
    notes.sort(key=attrgetter("time", "pitch", "duration", "velocity"))

    # Create the Track and Music objects
    track = Track(program=program, is_drum=is_drum, notes=notes)
    music = Music(resolution=resolution, tracks=[track])

    return music



"""Output interfaces.

This module provides output interfaces for common symbolic music
formats, MusPy's native JSON and YAML formats, other symbolic music
libraries and commonly-used representations in music generation.

Functions
---------

- save
- save_json
- save_yaml
- to_default_event_representation
- to_event_representation
- to_mido
- to_music21
- to_note_representation
- to_object
- to_performance_event_representation
- to_pianoroll_representation
- to_pitch_representation
- to_pretty_midi
- to_pypianoroll
- to_remi_event_representation
- to_representation
- write
-
- write_audio
- write_midi
- write_musicxml

"""
from .audio import synthesize, write_audio
from .event import (
    to_default_event_representation,
    to_event_representation,
    to_performance_event_representation,
    to_remi_event_representation,
)
from .json import save_json
from .midi import to_mido, to_pretty_midi, write_midi
from .music21 import to_music21
from .musicxml import write_musicxml
from .note import to_note_representation
from .pianoroll import to_pianoroll_representation, to_pypianoroll
from .pitch import to_pitch_representation
from .wrappers import save, to_object, to_representation, write
from .yaml import save_yaml

__all__ = [
    "save",
    "save_json",
    "save_yaml",
    "synthesize",
    "to_default_event_representation",
    "to_event_representation",
    "to_mido",
    "to_music21",
    "to_note_representation",
    "to_object",
    "to_performance_event_representation",
    "to_pianoroll_representation",
    "to_pitch_representation",
    "to_pretty_midi",
    "to_pypianoroll",
    "to_remi_event_representation",
    "to_representation",
    "write",
    "write_audio",
    "write_midi",
    "write_musicxml",
]



"""Visualization tools.

This module provides functions for visualizing a Music object.

Classes
-------

- ScorePlotter

Functions
---------

- show
- show_pianoroll
- show_score

"""
from .pianoroll import show_pianoroll
from .score import ScorePlotter, show_score
from .wrappers import show

__all__ = ["show", "show_pianoroll", "show_score", "ScorePlotter"]



"""Event-based representation input interface."""
from collections import defaultdict
from operator import attrgetter

import numpy as np
from numpy import ndarray

from ..classes import DEFAULT_VELOCITY, Note, Track
from ..music import DEFAULT_RESOLUTION, Music


def from_event_representation(
    array: ndarray,
    resolution: int = DEFAULT_RESOLUTION,
    program: int = 0,
    is_drum: bool = False,
    use_single_note_off_event: bool = False,
    use_end_of_sequence_event: bool = False,
    max_time_shift: int = 100,
    velocity_bins: int = 32,
    default_velocity: int = DEFAULT_VELOCITY,
    duplicate_note_mode: str = "fifo",
) -> Music:
    """Decode event-based representation into a Music object.

    Parameters
    ----------
    array : ndarray
        Array in event-based representation to decode.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.
    program : int, default: 0 (Acoustic Grand Piano)
        Program number, according to General MIDI specification [1].
        Valid values are 0 to 127.
    is_drum : bool, default: False
        Whether it is a percussion track.
    use_single_note_off_event : bool, default: False
        Whether to use a single note-off event for all the pitches. If
        True, a note-off event will close all active notes, which can
        lead to lossy conversion for polyphonic music.
    use_end_of_sequence_event : bool, default: False
        Whether to append an end-of-sequence event to the encoded
        sequence.
    max_time_shift : int, default: 100
        Maximum time shift (in ticks) to be encoded as an separate
        event. Time shifts larger than `max_time_shift` will be
        decomposed into two or more time-shift events.
    velocity_bins : int, default: 32
        Number of velocity bins to use.
    default_velocity : int, default: `muspy.DEFAULT_VELOCITY` (64)
        Default velocity value to use when decoding.
    duplicate_note_mode : {'fifo', 'lifo', 'all'}, default: 'fifo'
        Policy for dealing with duplicate notes. When a note off event
        is presetned while there are multiple correspoding note on
        events that have not yet been closed, we need a policy to decide
        which note on messages to close. This is only effective when
        `use_single_note_off_event` is False.

        - 'fifo' (first in first out): close the earliest note on
        - 'lifo' (first in first out): close the latest note on
        - 'all': close all note on messages

    Returns
    -------
    :class:`muspy.Music`
        Decoded Music object.

    References
    ----------
    [1] https://www.midi.org/specifications/item/gm-level-1-sound-set

    """
    if duplicate_note_mode.lower() not in ("fifo", "lifo", "all"):
        raise ValueError(
            "`duplicate_note_mode` must be one of 'fifo', 'lifo' and " "'all'."
        )

    # Cast the array to integer
    if not np.issubdtype(array.dtype, np.integer):
        raise TypeError("Array must be of type int.")

    # Compute offsets
    offset_note_on = 0
    offset_note_off = 128
    offset_time_shift = 129 if use_single_note_off_event else 256
    offset_velocity = offset_time_shift + max_time_shift
    if use_end_of_sequence_event:
        offset_eos = offset_velocity + velocity_bins

    # Compute vocabulary size
    if use_single_note_off_event:
        vocab_size = 129 + max_time_shift + velocity_bins
    else:
        vocab_size = 256 + max_time_shift + velocity_bins
    if use_end_of_sequence_event:
        vocab_size += 1

    # Decode events
    time = 0
    velocity = default_velocity
    velocity_factor = 128 / velocity_bins
    notes = []

    # Keep track of active note on messages
    active_notes = defaultdict(list)

    # Iterate over the events
    for event in array.flatten().tolist():
        # Skip unknown events
        if event < offset_note_on or event >= vocab_size:
            continue

        # End-of-sequence events
        if use_end_of_sequence_event and event == offset_eos:
            break

        # Note on events
        if event < offset_note_off:
            pitch = event - offset_note_on
            active_notes[pitch].append(
                Note(
                    time=int(time),
                    pitch=int(pitch),
                    duration=0,
                    velocity=int(velocity),
                )
            )

        # Note off events
        elif event < offset_time_shift:
            # Close all notes
            if use_single_note_off_event:
                if active_notes:
                    for pitch, note_list in active_notes.items():
                        for note in note_list:
                            note.duration = int(time - note.time)
                            notes.append(note)
                    active_notes = defaultdict(list)
                continue

            pitch = event - offset_note_off

            # Skip it if there is no active notes
            if not active_notes[pitch]:
                continue

            # NOTE: There is no way to disambiguate duplicate notes of
            # the same pitch. Thus, we need a policy for handling
            # duplicate notes.

            # 'FIFO': (first in first out) close the earliest note
            elif duplicate_note_mode.lower() == "fifo":
                note = active_notes[pitch][0]
                note.duration = int(time - note.time)
                notes.append(note)
                del active_notes[pitch][0]

            # 'LIFO': (last in first out) close the latest note on
            elif duplicate_note_mode.lower() == "lifo":
                note = active_notes[pitch][-1]
                note.duration = int(time - note.time)
                notes.append(note)
                del active_notes[pitch][-1]

            # 'all' - close all note on events
            elif duplicate_note_mode.lower() == "all":
                for note in active_notes[pitch]:
                    note.duration = int(time - note.time)
                    notes.append(note)
                del active_notes[pitch]

        # Time-shift events
        elif event < offset_velocity:
            time += event - offset_time_shift + 1

        # Velocity events
        elif event < vocab_size:
            velocity = int((event - offset_velocity) * velocity_factor)

    # Sort the notes
    notes.sort(key=attrgetter("time", "pitch", "duration", "velocity"))

    # Create the Track and Music objects
    track = Track(program=program, is_drum=is_drum, notes=notes)
    music = Music(resolution=resolution, tracks=[track])

    return music



"""MusicNet Dataset."""
from pathlib import Path
from typing import Union

from ..inputs import read_midi
from ..music import Music
from .base import DatasetInfo, RemoteFolderDataset

# pylint: disable=line-too-long

_NAME = "MusicNet Dataset"
_DESCRIPTION = """\
MusicNet is a collection of 330 freely-licensed classical music recordings, \
together with over 1 million annotated labels indicating the precise time of \
each note in every recording, the instrument that plays each note, and the \
note's position in the metrical structure of the composition."""
_HOMEPAGE = "https://homes.cs.washington.edu/~thickstn/musicnet.html"
_CITATION = """\
@inproceedings{thickstun2017learning,
  author={John Thickstun and Zaid Harchaoui and Sham M. Kakade},
  title={Learning Features of Music from Scratch},
  booktitle={International Conference on Learning Representations (ICLR)},
  year=2017
}"""


class MusicNetDataset(RemoteFolderDataset):
    """MusicNet Dataset (MIDI only)."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _citation = _CITATION
    _sources = {
        "musicnet": {
            "filename": "musicnet_midis.tar.gz",
            "url": "https://homes.cs.washington.edu/~thickstn/media/musicnet_midis.tar.gz",
            "archive": True,
            "size": 2601302,
            "md5": "b5fa98a113bfc51c8a445def9f24dc7e",
            "sha256": "943cc47731ec5f397bd6fbab4dff78342472890cd484bd30ffb2e16047eef908",
        }
    }
    _extension = "mid"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)



"""Utility functions for working with schemas."""
import json
from pathlib import Path
from typing import Union

try:
    import jsonschema

    _HAS_JSONSCHEMA = True
except ImportError:
    _HAS_JSONSCHEMA = False

try:
    import xmlschema

    _HAS_XMLSCHEMA = True
except ImportError:
    _HAS_XMLSCHEMA = False

try:
    import yamale

    _HAS_YAMALE = True
except ImportError:
    _HAS_YAMALE = False


def get_json_schema_path() -> str:
    """Return the path to the JSON schema."""
    return str(Path(__file__).resolve().parent / "music.schema.json")


def get_yaml_schema_path() -> str:
    """Return the path to the YAML schema."""
    return str(Path(__file__).resolve().parent / "music.schema.yaml")


def get_musicxml_schema_path() -> str:
    """Return the path to the MusicXML schema."""
    return str(Path(__file__).resolve().parent / "musicxml.xsd")


def validate_json(path: Union[str, Path]):
    """Validate a file against the JSON schema.

    Parameters
    ----------
    path : str or Path
        Path to the file to validate.

    """
    if not _HAS_JSONSCHEMA:
        raise RuntimeError(
            "The jsonschema library is required for JSON schema validation. "
            "You could install it by `pip install muspy[schema]`."
        )
    with open(str(path)) as f:
        data = json.load(f)
    with open(str(get_json_schema_path())) as f:
        schema = json.load(f)
    jsonschema.validate(data, schema)


def validate_yaml(path: Union[str, Path]):
    """Validate a file against the YAML schema.

    Parameters
    ----------
    path : str or Path
        Path to the file to validate.

    """
    if not _HAS_YAMALE:
        raise RuntimeError(
            "The Yamale library is required for YAML schema validation. "
            "You could install it by `pip install muspy[schema]`."
        )
    data = yamale.make_data(str(path))
    schema = yamale.make_schema(str(get_yaml_schema_path()))
    yamale.validate(schema, data)


def validate_musicxml(path: Union[str, Path]):
    """Validate a file against the MusicXML schema.

    Parameters
    ----------
    path : str or Path
        Path to the file to validate.

    """
    if not _HAS_XMLSCHEMA:
        raise RuntimeError(
            "The xmlschema library is required for MusicXML schema "
            "validation. You could install it by `pip install muspy[schema]`."
        )
    schema = xmlschema.XMLSchema(get_musicxml_schema_path())
    schema.validate(str(path))



"""A toolkit for symbolic music generation.

MusPy is an open source Python library for symbolic music generation. It
provides essential tools for developing a music generation system,
including dataset management, data I/O, data preprocessing and model
evaluation.

Features
--------

- Dataset management system for commonly used datasets with interfaces
  to PyTorch and TensorFlow.
- Data I/O for common symbolic music formats (e.g., MIDI, MusicXML and
  ABC) and interfaces to other symbolic music libraries (e.g., music21,
  mido, pretty_midi and Pypianoroll).
- Implementations of common music representations for music generation,
  including the pitch-based, the event-based, the piano-roll and the
  note-based representations.
- Model evaluation tools for music generation systems, including audio
  rendering, score and piano-roll visualizations and objective metrics.

"""
from . import (
    base,
    classes,
    core,
    datasets,
    external,
    inputs,
    metrics,
    music,
    outputs,
    processors,
    schemas,
    visualization,
)
from .base import *
from .classes import *
from .core import *
from .datasets import *
from .external import *
from .inputs import *
from .metrics import *
from .music import *
from .outputs import *
from .processors import *
from .schemas import *
from .version import __version__
from .visualization import *

__all__ = [
    "__version__",
    "datasets",
    "external",
    "inputs",
    "metrics",
    "outputs",
    "processors",
    "schemas",
    "visualization",
]
__all__.extend(base.__all__)
__all__.extend(classes.__all__)
__all__.extend(core.__all__)
__all__.extend(datasets.__all__)
__all__.extend(external.__all__)
__all__.extend(inputs.__all__)
__all__.extend(metrics.__all__)
__all__.extend(music.__all__)
__all__.extend(outputs.__all__)
__all__.extend(processors.__all__)
__all__.extend(schemas.__all__)
__all__.extend(visualization.__all__)



"""MIDI input interface."""
import warnings
from collections import OrderedDict, defaultdict
from operator import attrgetter
from pathlib import Path
from typing import List, Union

import numpy as np
from mido import MidiFile, tempo2bpm
from pretty_midi import Instrument
from pretty_midi import KeySignature as PmKeySignature
from pretty_midi import Lyric as PmLyric
from pretty_midi import Note as PmNote
from pretty_midi import PrettyMIDI
from pretty_midi import TimeSignature as PmTimeSignature

from ..classes import (
    Annotation,
    KeySignature,
    Lyric,
    Metadata,
    Note,
    Tempo,
    TimeSignature,
    Track,
)
from ..music import DEFAULT_RESOLUTION, Music
from ..utils import note_str_to_note_num


class MIDIError(Exception):
    """An error class for MIDI related exceptions."""


def _is_drum(channel):
    # Mido numbers channels 0 to 15 instead of 1 to 16
    return channel == 9


def from_mido(midi: MidiFile, duplicate_note_mode: str = "fifo") -> Music:
    """Return a mido MidiFile object as a Music object.

    Parameters
    ----------
    midi : :class:`mido.MidiFile`
        Mido MidiFile object to convert.
    duplicate_note_mode : {'fifo', 'lifo', 'all'}, default: 'fifo'
        Policy for dealing with duplicate notes. When a note off
        message is presetned while there are multiple correspoding note
        on messages that have not yet been closed, we need a policy to
        decide which note on messages to close.

        - 'fifo' (first in first out): close the earliest note on
        - 'lifo' (first in first out): close the latest note on
        - 'all': close all note on messages

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    if duplicate_note_mode.lower() not in ("fifo", "lifo", "all"):
        raise ValueError(
            "`duplicate_note_mode` must be one of 'fifo', 'lifo' and " "'all'."
        )

    def _get_active_track(t_idx, program, channel):
        """Return the active track."""
        key = (program, channel)
        if key in tracks[t_idx]:
            return tracks[t_idx][key]
        tracks[t_idx][key] = Track(program, _is_drum(channel))
        return tracks[t_idx][key]

    # Raise MIDIError if the MIDI file is of Type 2 (i.e., asynchronous)
    if midi.type == 2:
        raise MIDIError("Type 2 MIDI file is not supported.")

    # Raise MIDIError if ticks_per_beat is not positive
    if midi.ticks_per_beat < 1:
        raise MIDIError("`ticks_per_beat` must be positive.")

    time = 0
    song_title = None
    tempos: List[Tempo] = []
    key_signatures: List[KeySignature] = []
    time_signatures: List[TimeSignature] = []
    lyrics: List[Lyric] = []
    annotations: List[Annotation] = []
    copyrights = []

    # Create a list to store converted tracks
    tracks: List[OrderedDict] = [
        OrderedDict() for _ in range(len(midi.tracks))
    ]
    # Create a list to store track names
    track_names = [None] * len(midi.tracks)

    # Iterate over MIDI tracks
    for track_idx, midi_track in enumerate(midi.tracks):

        # Set current time to zero
        time = 0
        # Keep track of the program used in each channel
        channel_programs = [0] * 16
        # Keep track of active note on messages
        active_notes = defaultdict(list)

        # Iterate over MIDI messages
        for msg in midi_track:

            # Update current time (delta time is used in a MIDI message)
            time += msg.time

            # === Meta Data ===

            # Tempo messages
            if msg.type == "set_tempo":
                tempos.append(
                    Tempo(time=int(time), qpm=float(tempo2bpm(msg.tempo)))
                )

            # Key signature messages
            elif msg.type == "key_signature":
                if msg.key.endswith("m"):
                    mode = "minor"
                    root = note_str_to_note_num(msg.key[:-1])
                else:
                    mode = "major"
                    root = note_str_to_note_num(msg.key)
                key_signatures.append(
                    KeySignature(time=int(time), root=root, mode=mode)
                )

            # Time signature messages
            elif msg.type == "time_signature":
                time_signatures.append(
                    TimeSignature(
                        time=int(time),
                        numerator=int(msg.numerator),
                        denominator=int(msg.denominator),
                    )
                )

            # Lyric messages
            elif msg.type == "lyrics":
                lyrics.append(Lyric(time=int(time), lyric=str(msg.text)))

            # Marker messages
            elif msg.type == "marker":
                annotations.append(
                    Annotation(
                        time=int(time),
                        annotation=str(msg.text),
                        group="marker",
                    )
                )

            # Text messages
            elif msg.type == "text":
                annotations.append(
                    Annotation(
                        time=int(time), annotation=str(msg.text), group="text"
                    )
                )

            # Copyright messages
            elif msg.type == "copyright":
                copyrights.append(str(msg.text))

            # === Track specific Data ===

            # Track name messages
            elif msg.type == "track_name":
                if midi.type == 0 or track_idx == 0:
                    song_title = msg.name
                else:
                    track_names[track_idx] = msg.name

            # Program change messages
            elif msg.type == "program_change":
                # Change program of the channel
                channel_programs[msg.channel] = msg.program

            # Note on messages
            elif msg.type == "note_on" and msg.velocity > 0:
                # Will later be closed by a note off message
                active_notes[(msg.channel, msg.note)].append(
                    (time, msg.velocity)
                )

            # Note off messages
            # NOTE: A note on message with a zero velocity is also
            # considered a note off message
            elif msg.type == "note_off" or (
                msg.type == "note_on" and msg.velocity == 0
            ):
                # Skip it if there is no active notes
                note_key = (msg.channel, msg.note)
                if not active_notes[note_key]:
                    continue

                # Get the active track
                program = channel_programs[msg.channel]
                track = _get_active_track(track_idx, program, msg.channel)

                # NOTE: There is no way to disambiguate duplicate notes
                # (of the same pitch on the same channel). Thus, we
                # need a policy for handling duplicate notes.

                # 'FIFO': (first in first out) close the earliest note
                if duplicate_note_mode.lower() == "fifo":
                    onset, velocity = active_notes[note_key][0]
                    track.notes.append(
                        Note(
                            time=int(onset),
                            pitch=int(msg.note),
                            duration=int(time - onset),
                            velocity=int(velocity),
                        )
                    )
                    del active_notes[note_key][0]

                # 'LIFO': (last in first out) close the latest note on
                elif duplicate_note_mode.lower() == "lifo":
                    onset, velocity = active_notes[note_key][-1]
                    track.notes.append(
                        Note(
                            time=int(onset),
                            pitch=int(msg.note),
                            duration=int(time - onset),
                            velocity=int(velocity),
                        )
                    )
                    del active_notes[note_key][-1]

                # 'close_all' - close all note on messages
                elif duplicate_note_mode.lower() == "close_all":
                    for onset, velocity in active_notes[note_key]:
                        track.notes.append(
                            Note(
                                time=int(onset),
                                pitch=int(msg.note),
                                duration=int(time - onset),
                                velocity=int(velocity),
                            )
                        )
                    del active_notes[note_key]

            # Control change messages
            elif msg.type == "control_change":
                # Get the active track
                program = channel_programs[msg.channel]
                track = _get_active_track(track_idx, program, msg.channel)

                # Append the control change message as an annotation
                track.annotations.append(
                    Annotation(
                        time=int(time),
                        annotation={
                            "number": int(msg.control),
                            "value": int(msg.value),
                        },
                        group="control_change",
                    )
                )

            # End of track message
            elif msg.type == "end_of_track":
                break

        # Close all active notes
        for (channel, note), note_ons in active_notes.items():
            program = channel_programs[channel]
            track = _get_active_track(track_idx, program, channel)
            for onset, velocity in note_ons:
                track.notes.append(
                    Note(
                        time=int(onset),
                        pitch=int(note),
                        duration=int(time - onset),
                        velocity=int(velocity),
                    )
                )

    music_tracks = []
    for track, track_name in zip(tracks, track_names):
        for sub_track in track.values():
            sub_track.name = track_name
        music_tracks.extend(track.values())

    # Sort notes
    for music_track in music_tracks:
        music_track.notes.sort(
            key=attrgetter("time", "pitch", "duration", "velocity")
        )

    # Meta data
    metadata = Metadata(
        title=str(song_title),
        source_format="midi",
        copyright=" ".join(copyrights) if copyrights else None,
    )

    return Music(
        metadata=metadata,
        resolution=int(midi.ticks_per_beat),
        tempos=tempos,
        key_signatures=key_signatures,
        time_signatures=time_signatures,
        lyrics=lyrics,
        tracks=music_tracks,
    )


def read_midi_mido(
    path: Union[str, Path], duplicate_note_mode: str = "fifo"
) -> Music:
    """Read a MIDI file into a Music object using mido backend.

    Parameters
    ----------
    path : str or Path
        Path to the MIDI file to read.
    duplicate_note_mode : {'fifo', 'lifo, 'all'}, default: 'fifo'
        Policy for dealing with duplicate notes. When a note off message
        is presetned while there are multiple correspoding note on
        messages that have not yet been closed, we need a policy to
        decide which note on messages to close.

        - 'fifo' (first in first out): close the earliest note on
        - 'lifo' (first in first out):close the latest note on
        - 'all': close all note on messages

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    midi = MidiFile(filename=str(path))
    music = from_mido(midi, duplicate_note_mode=duplicate_note_mode)
    music.metadata.source_filename = Path(path).name
    return music


def from_pretty_midi_key_signature(
    key_signature: PmKeySignature,
) -> KeySignature:
    """Return a pretty_midi KeySignature object as a KeySignature.

    Parameters
    ----------
    key_signature : :class:`pretty_midi.KeySignature`
        pretty_midi KeySignature object to convert.

    Returns
    -------
    :class:`muspy.KeySignature`
        Converted key signature.

    Note
    ----
    The `time` attribute of the converted object will be of type float
    as pretty_midi uses the absolute timing system.

    """
    is_minor, root = divmod(key_signature.key_number, 12)
    mode = "minor" if is_minor else "major"
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        return KeySignature(
            time=float(key_signature.time),  # type: ignore
            root=root,
            mode=mode,
        )


def from_pretty_midi_time_signature(
    time_signature: PmTimeSignature,
) -> TimeSignature:
    """Return a pretty_midi TimeSignature object as a TimeSignature.

    Parameters
    ----------
    time_signature : :class:`pretty_midi.TimeSignature`
        pretty_midi TimeSignature object to convert.

    Returns
    -------
    :class:`muspy.TimeSignature`
        Converted time signature.

    Note
    ----
    The `time` attribute of the converted object will be of type float
    as pretty_midi uses the absolute timing system.

    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        return TimeSignature(
            time=float(time_signature.time),  # type: ignore
            numerator=time_signature.numerator,
            denominator=time_signature.denominator,
        )


def from_pretty_midi_lyric(lyric: PmLyric) -> Lyric:
    """Return a pretty_midi Lyric object as a Lyric object.

    Parameters
    ----------
    lyric : :class:`pretty_midi.Lyric`
        pretty_midi Lyric object to convert.

    Returns
    -------
    :class:`muspy.Lyric`
        Converted lyric.

    Note
    ----
    The `time` attribute of the converted object will be of type float
    as pretty_midi uses the absolute timing system.

    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        return Lyric(
            time=float(lyric.time),  # type: ignore
            lyric=str(lyric.text),
        )


def from_pretty_midi_note(note: PmNote) -> Note:
    """Return pretty_midi Note object as a Note object.

    Parameters
    ----------
    note : :class:`pretty_midi.Note`
        pretty_midi Note object to convert.

    Returns
    -------
    :class:`muspy.Note`
        Converted note.

    Note
    ----
    The `time` and `duration` attributes of the converted object will be
    of type float as pretty_midi uses the absolute timing system.

    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        return Note(
            time=float(note.start),  # type: ignore
            duration=float(note.duration),  # type: ignore
            pitch=int(note.pitch),
            velocity=int(note.velocity),
        )


def from_pretty_midi_instrument(instrument: Instrument) -> Track:
    """Return a pretty_midi Instrument object as a Track object.

    Parameters
    ----------
    instrument : :class:`pretty_midi.Instrument`
        pretty_midi Instrument object to convert.

    Returns
    -------
    :class:`muspy.Track`
        Converted track.

    """
    return Track(
        program=int(instrument.program),
        is_drum=bool(instrument.is_drum),
        name=str(instrument.name),
        notes=[from_pretty_midi_note(note) for note in instrument.notes],
    )


def from_pretty_midi(midi: PrettyMIDI, resolution: int = None) -> Music:
    """Return a pretty_midi PrettyMIDI object as a Music object.

    Parameters
    ----------
    midi : :class:`pretty_midi.PrettyMIDI`
        PrettyMIDI object to convert.
    resolution : int, default: `muspy.DEFAULT_RESOLUTION` (24)
        Time steps per quarter note.

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    if resolution is None:
        resolution = DEFAULT_RESOLUTION

    tempo_realtimes, tempi = midi.get_tempo_changes()
    assert len(tempi) > 0
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        tempos = [
            Tempo(time=float(time), qpm=float(tempo))  # type: ignore
            for time, tempo in zip(tempo_realtimes, tempi)
        ]

    key_signatures = [
        from_pretty_midi_key_signature(key_signature)
        for key_signature in midi.key_signature_changes
    ]
    time_signatures = [
        from_pretty_midi_time_signature(time_signature)
        for time_signature in midi.time_signature_changes
    ]
    lyrics = [from_pretty_midi_lyric(lyric) for lyric in midi.lyrics]
    tracks = [from_pretty_midi_instrument(track) for track in midi.instruments]
    music = Music(
        metadata=Metadata(source_format="midi"),
        tempos=tempos,
        key_signatures=key_signatures,
        time_signatures=time_signatures,
        lyrics=lyrics,
        tracks=tracks,
    )

    # NOTE: pretty_midi uses the absolute timing system, so we have to
    # convert all the timings into metrical timing.

    # Remove unnecessary tempo changes to speed up the search
    if len(tempi) > 1:
        last_tempo = tempi[0]
        last_time = tempo_realtimes[0]
        tempo_realtimes = tempo_realtimes.tolist()
        tempi = tempi.tolist()
        i = 1
        while i < len(tempo_realtimes):
            if tempi[i] == last_tempo:
                del tempo_realtimes[i]
                del tempi[i]
            elif tempo_realtimes[i] == last_time:
                del tempo_realtimes[i - 1]
                del tempi[i - 1]
            else:
                last_tempo = tempi[i]
                i += 1
        tempo_realtimes = np.array(tempo_realtimes)
        tempi = np.array(tempi)

    if len(tempi) == 1:

        def map_time(time: float) -> int:
            factor = resolution * float(tempi[0]) / 60.0  # type: ignore
            return round(time * factor)

    else:
        # Compute the tempo time in metrical timing of each tempo change
        tempo_times = np.cumsum(
            np.diff(tempo_realtimes) * resolution * tempi[:-1] / 60.0
        )
        tempo_times = np.round(tempo_times).astype(int).tolist()
        tempo_times.insert(0, 0)

        def map_time(time: float) -> int:
            idx = np.searchsorted(tempo_realtimes, time, side="right") - 1
            residual = time - tempo_realtimes[idx]
            factor = resolution * tempi[idx] / 60.0
            return round(tempo_times[idx] + residual * factor)

    # Adjust timing
    music.adjust_time(func=map_time)

    return music


def read_midi_pretty_midi(path: Union[str, Path]) -> Music:
    """Read a MIDI file into a Music object using pretty_midi backend.

    Parameters
    ----------
    path : str or Path
        Path to the MIDI file to read.

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    music = from_pretty_midi(PrettyMIDI(str(path)))
    music.metadata.source_filename = Path(path).name
    return music


def read_midi(
    path: Union[str, Path],
    backend: str = "mido",
    duplicate_note_mode: str = "fifo",
) -> Music:
    """Read a MIDI file into a Music object.

    Parameters
    ----------
    path : str or Path
        Path to the MIDI file to read.
    backend: {'mido', 'pretty_midi'}, default: 'mido'
        Backend to use.
    duplicate_note_mode : {'fifo', 'lifo, 'all'}, default: 'fifo'
        Policy for dealing with duplicate notes. When a note off message
        is presetned while there are multiple correspoding note on
        messages that have not yet been closed, we need a policy to
        decide which note on messages to close. Only used when `backend`
        is 'mido'.

        - 'fifo' (first in first out): close the earliest note on
        - 'lifo' (first in first out):close the latest note on
        - 'all': close all note on messages

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    """
    if backend == "mido":
        return read_midi_mido(path, duplicate_note_mode=duplicate_note_mode)
    if backend == "pretty_midi":
        return read_midi_pretty_midi(path)
    raise ValueError("`backend` must by one of 'mido' and 'pretty_midi'.")



"""JSB Chorales Dataset."""
from pathlib import Path
from typing import Union

from ..inputs import read_midi
from ..music import Music
from .base import DatasetInfo, RemoteFolderDataset

# pylint: disable=line-too-long

_NAME = "JSB Chorales Dataset"
_DESCRIPTION = """\
The JSB Chorales Dataset is a collection of 382 four-part chorales by Johann \
Sebastian Bach. This dataset is used in the paper "Modeling Temporal \
Dependencies in High-Dimensional Sequences: Application to Polyphonic Music \
Generation and Transcription" in ICML 2012. It comes with train, test and \
validation split used in the paper "Harmonising Chorales by Probabilistic \
Inference" in NIPS 2005."""
_HOMEPAGE = "http://www-etud.iro.umontreal.ca/~boulanni/icml2012"
_CITATION = """\
@inproceedings{boulangerlewandowski2012modeling,
  author={Nicolas Boulanger-Lewandowski and Yoshua Bengio and Pascal Vincent},
  title={Modeling Temporal Dependencies in High-Dimensional Sequences: Application to Polyphonic Music Generation and Transcription},
  booktitle={Proceedings of the 29th International Conference on Machine Learning (ICML)},
  year=2012
}"""


class JSBChoralesDataset(RemoteFolderDataset):
    """Johann Sebastian Bach Chorales Dataset."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _citation = _CITATION
    _sources = {
        "jsb": {
            "filename": "JSB Chorales.zip",
            "url": "http://www-etud.iro.umontreal.ca/~boulanni/JSB%20Chorales.zip",
            "archive": True,
            "size": 215242,
            "md5": "2fb72faf2659e82e9de08b16f2cca1e9",
            "sha256": "69924294546a71620c06374085cd6b0300665ea215e2f854f65a11929f1e723c",
        }
    }
    _extension = "mid"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        music = read_midi(self.root / filename)

        # The resolution of MIDI file in this datset should be 120, but
        # is incorrectly set to 100
        music.resolution = 120

        return music



"""NES Music Database."""
from pathlib import Path
from typing import Union

from ..inputs import read_midi
from ..music import Music
from .base import DatasetInfo, RemoteFolderDataset

# pylint: disable=line-too-long

_NAME = "NES Music Database"
_DESCRIPTION = """\
The Nintendo Entertainment System Music Database (NES-MDB) is a dataset \
intended for building automatic music composition systems for the NES audio \
synthesizer. The NES synthesizer has highly constrained compositional \
parameters which are well-suited to a wide variety of current machine \
learning techniques. The synthesizer is typically programmed in assembly, but \
we parse the assembly into straightforward formats that are more suitable for \
machine learning."""
_HOMEPAGE = "https://github.com/chrisdonahue/nesmdb"
_CITATION = """\
@inproceedings{donahue2018nesmdb,
  author={Chris Donahue and Huanru Henry Mao and Julian McAuley},
  title={The {NES} Music Database: A multi-instrumental dataset with expressive performance attributes},
  booktitle={Proceedings of the 19th International Society for Music Retrieval Conference (ISMIR)},
  year=2018
}"""


class NESMusicDatabase(RemoteFolderDataset):
    """NES Music Database."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _citation = _CITATION
    _sources = {
        "nes": {
            "filename": "nesmdb_midi.tar.gz",
            "url": "http://deepyeti.ucsd.edu/cdonahue/nesmdb/nesmdb_midi.tar.gz",
            "archive": True,
            "size": 12922275,
            "md5": "3f3e8ab4f660dd1b19350e5a8a91f3e6",
            "sha256": "37610e2ca5fe70359c85170cf1f4982596783bb304c59d9c439f68c24ff4ee93",
        }
    }
    _extension = "mid"

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)



"""Music21 converter interface."""
from typing import TYPE_CHECKING

from music21.key import Key
from music21.metadata import Contributor, Copyright
from music21.metadata import Metadata as M21MetaData
from music21.meter import TimeSignature as M21TimeSignature
from music21.note import Note as M21Note
from music21.stream import Part, Score
from music21.tempo import MetronomeMark

from ..classes import KeySignature, Metadata, Tempo, TimeSignature
from ..utils import CIRCLE_OF_FIFTHS, MODE_CENTERS

if TYPE_CHECKING:
    from ..music import Music

PITCH_NAMES = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]


def _get_pitch_name(note_number: int) -> str:
    octave, pitch_class = divmod(note_number, 12)
    return PITCH_NAMES[pitch_class] + str(octave - 1)


def to_music21_metronome(tempo: Tempo) -> MetronomeMark:
    """Return a Tempo object as a music21 MetronomeMark object."""
    metronome = MetronomeMark(number=tempo.qpm)
    metronome.offset = tempo.time
    return metronome


def to_music21_key(key_signature: KeySignature) -> Key:
    """Return a KeySignature object as a music21 Key object."""
    if key_signature.root_str is not None:
        tonic = key_signature.root_str
    elif key_signature.root is not None:
        tonic = PITCH_NAMES[key_signature.root]
    elif key_signature.fifths is not None:
        if key_signature.mode is not None:
            offset = MODE_CENTERS[key_signature.mode]
            tonic = CIRCLE_OF_FIFTHS[key_signature.fifths + offset][1]
        else:
            tonic = CIRCLE_OF_FIFTHS[key_signature.fifths][1]
    else:
        raise ValueError(
            "One of `root`, `root_str` or `fifths` must be specified."
        )
    key = Key(tonic=tonic, mode=key_signature.mode)
    key.offset = key_signature.time
    return key


def to_music21_time_signature(
    time_signature: TimeSignature,
) -> M21TimeSignature:
    """Return a TimeSignature object as a music21 TimeSignature."""
    m21_time_signature = M21TimeSignature(
        "{}/{}".format(time_signature.numerator, time_signature.denominator)
    )
    m21_time_signature.offset = time_signature.time
    return m21_time_signature


def to_music21_metadata(metadata: Metadata) -> M21MetaData:
    """Return a Metadata object as a music21 Metadata object.

    Parameters
    ----------
    metadata : :class:`muspy.Metadata`
        Metadata object to convert.

    Returns
    -------
    `music21.metadata.Metadata`
        Converted music21 Metadata object.

    """
    meta = M21MetaData()

    # Title is usually stored in movement-title. See
    # https://www.musicxml.com/tutorial/file-structure/score-header-entity/
    if metadata.title:
        meta.movementName = metadata.title

    if metadata.copyright:
        meta.copyright = Copyright(metadata.copyright)
    for creator in metadata.creators:
        meta.addContributor(Contributor(name=creator))
    return meta


def to_music21(music: "Music") -> Score:
    """Return a Music object as a music21 Score object.

    Parameters
    ----------
    music : :class:`muspy.Music`
        Music object to convert.

    Returns
    -------
    `music21.stream.Score`
        Converted music21 Score object.

    """
    # Create a new score
    score = Score()

    # Metadata
    if music.metadata:
        score.append(to_music21_metadata(music.metadata))

    # Tracks
    for track in music.tracks:
        # Create a new part
        part = Part()
        part.partName = track.name

        # Add tempos
        for tempo in music.tempos:
            part.append(to_music21_metronome(tempo))

        # Add time signatures
        for time_signature in music.time_signatures:
            part.append(to_music21_time_signature(time_signature))

        # Add key signatures
        for key_signature in music.key_signatures:
            part.append(to_music21_key(key_signature))

        # Add notes to part
        for note in track.notes:
            m21_note = M21Note(_get_pitch_name(note.pitch))
            m21_note.offset = note.time / music.resolution
            m21_note.quarterLength = note.duration / music.resolution
            part.append(m21_note)

        # Append the part to score
        score.append(part)

    return score



"""Core classes.

This module defines the core classes of MusPy.

Classes
-------

- Annotation
- Beat
- Chord
- KeySignature
- Lyric
- Metadata
- Note
- Tempo
- TimeSignature
- Track


Variables
---------

- DEFAULT_VELOCITY

"""
from collections import OrderedDict
from typing import Any, Callable, List

from .base import Base, ComplexBase
from .schemas import DEFAULT_SCHEMA_VERSION

DEFAULT_VELOCITY = 64

__all__ = [
    "Annotation",
    "Beat",
    "Chord",
    "DEFAULT_VELOCITY",
    "KeySignature",
    "Lyric",
    "Metadata",
    "Note",
    "Tempo",
    "TimeSignature",
    "Track",
]

# pylint: disable=super-init-not-called


class Metadata(Base):
    """A container for metadata.

    Attributes
    ----------
    schema_version : str, default: `muspy.DEFAULT_SCHEMA_VERSION`
        Schema version.
    title : str, optional
        Song title.
    creators : list of str, optional
        Creator(s) of the song.
    copyright : str, optional
        Copyright notice.
    collection : str, optional
        Name of the collection.
    source_filename : str, optional
        Name of the source file.
    source_format : str, optional
        Format of the source file.

    """

    _attributes = OrderedDict(
        [
            ("schema_version", str),
            ("title", str),
            ("creators", str),
            ("copyright", str),
            ("collection", str),
            ("source_filename", str),
            ("source_format", str),
        ]
    )
    _optional_attributes = [
        "title",
        "creators",
        "copyright",
        "collection",
        "source_filename",
        "source_format",
    ]
    _list_attributes = ["creators"]

    def __init__(
        self,
        schema_version: str = DEFAULT_SCHEMA_VERSION,
        title: str = None,
        creators: List[str] = None,
        copyright: str = None,
        collection: str = None,
        source_filename: str = None,
        source_format: str = None,
    ):
        # pylint: disable=redefined-builtin
        self.schema_version = schema_version
        self.title = title
        self.creators = creators if creators is not None else []
        self.copyright = copyright
        self.collection = collection
        self.source_filename = source_filename
        self.source_format = source_format


class Tempo(Base):
    """A container for key signatures.

    Attributes
    ----------
    time : int
        Start time of the tempo, in time steps.
    qpm : float
        Tempo in qpm (quarters per minute).

    """

    _attributes = OrderedDict([("time", int), ("qpm", (float, int))])

    def __init__(self, time: int, qpm: float):
        self.time = time
        self.qpm = float(qpm)

    def _validate(self, attr: str, recursive: bool):
        super()._validate(attr, recursive)
        if attr == "qpm" and self.qpm <= 0:
            raise ValueError("`qpm` must be positive.")


class KeySignature(Base):
    """A container for key signatures.

    Attributes
    ----------
    time : int
        Start time of the key signature, in time steps.
    root : int, optional
        Root (tonic) of the key signature.
    mode : str, optional
        Mode of the key signature.
    fifths : int, optional
        Number of sharps or flats. Positive numbers for sharps and
        negative numbers for flats.
    root_str : str, optional
        Root of the key signature as a string.

    Note
    ----
    A key signature can be specified either by its root (`root`) or the
    number of sharps or flats (`fifths`) along with its mode.

    """

    _attributes = OrderedDict(
        [
            ("time", int),
            ("root", int),
            ("mode", str),
            ("fifths", int),
            ("root_str", str),
        ]
    )
    _optional_attributes = ["root", "mode", "fifths", "root_str"]

    def __init__(
        self,
        time: int,
        root: int = None,
        mode: str = None,
        fifths: int = None,
        root_str: str = None,
    ):
        self.time = time
        self.root = root
        self.mode = mode
        self.fifths = fifths
        self.root_str = root_str


class TimeSignature(Base):
    """A container for time signatures.

    Attributes
    ----------
    time : int
        Start time of the time signature, in time steps.
    numerator : int
        Numerator of the time signature.
    denominator : int
        Denominator of the time signature.

    """

    _attributes = OrderedDict(
        [("time", int), ("numerator", int), ("denominator", int)]
    )

    def __init__(self, time: int, numerator: int, denominator: int):
        self.time = time
        self.numerator = numerator
        self.denominator = denominator

    def _validate(self, attr: str, recursive: bool):
        super()._validate(attr, recursive)
        if attr == "numerator" and self.numerator < 1:
            raise ValueError("`numerator` must be positive.")
        if attr == "denominator" and self.denominator < 1:
            raise ValueError("`denominator` must be positive.")


class Beat(Base):
    """A container for beats.

    Attributes
    ----------
    time : int
        Time of the beat, in time steps.
    is_downbeat : bool, default: False
        Whether it is a downbeat.

    """

    _attributes = OrderedDict([("time", int), ("is_downbeat", bool)])
    _optional_attributes = ["is_downbeat"]

    def __init__(self, time: int, is_downbeat: bool = False):
        self.time = time
        self.is_downbeat = is_downbeat


class Lyric(Base):
    """A container for lyrics.

    Attributes
    ----------
    time : int
        Start time of the lyric, in time steps.
    lyric : str
        Lyric (sentence, word, syllable, etc.).

    """

    _attributes = OrderedDict([("time", int), ("lyric", str)])

    def __init__(self, time: int, lyric: str):
        self.time = time
        self.lyric = lyric


class Annotation(Base):
    """A container for annotations.

    Attributes
    ----------
    time : int
        Start time of the annotation, in time steps.
    annotation : any
        Annotation of any type.
    group : str, optional
        Group name (for better organizing the annotations).

    """

    _attributes = OrderedDict(
        [("time", int), ("annotation", object), ("group", str)]
    )
    _optional_attributes = ["group"]

    def __init__(self, time: int, annotation: Any, group: str = None):
        self.time = time
        self.annotation = annotation
        self.group = group


class Note(Base):
    """A container for notes.

    Attributes
    ----------
    time : int
        Start time of the note, in time steps.
    pitch : int
        Note pitch, as a MIDI note number. Valid values are 0 to 127.
    duration : int
        Duration of the note, in time steps.
    velocity : int, default: `muspy.DEFAULT_VELOCITY` (64)
        Note velocity. Valid values are 0 to 127.
    pitch_str : str, optional
        Note pitch as a string, useful for distinguishing, e.g., C# and
        Db.

    """

    _attributes = OrderedDict(
        [
            ("time", int),
            ("pitch", int),
            ("duration", int),
            ("velocity", int),
            ("pitch_str", str),
        ]
    )
    _optional_attributes = ["velocity", "pitch_str"]

    def __init__(
        self,
        time: int,
        pitch: int,
        duration: int,
        velocity: int = None,
        pitch_str: str = None,
    ):
        self.time = time
        self.pitch = pitch
        self.duration = duration
        self.velocity = velocity if velocity is not None else DEFAULT_VELOCITY
        self.pitch_str = pitch_str

    @property
    def start(self):
        """Start time of the note."""
        return self.time

    @start.setter
    def start(self, start):
        """Setter for start time."""
        self.time = start

    @property
    def end(self):
        """End time of the note."""
        return self.time + self.duration

    @end.setter
    def end(self, end):
        """Setter for end time."""
        self.duration = end - self.time

    def _validate(self, attr: str, recursive: bool):
        super()._validate(attr, recursive)
        if attr == "pitch" and (self.pitch < 0 or self.pitch > 127):
            raise ValueError("`pitch` must be in between 0 to 127.")
        if attr == "duration" and self.duration < 0:
            raise ValueError("`duration` must be nonnegative.")
        if attr == "velocity" and (self.velocity < 0 or self.velocity > 127):
            raise ValueError("`velocity` must be in between 0 to 127.")

    def _adjust_time(
        self, func: Callable[[int], int], attr: str, recursive: bool
    ):
        raise NotImplementedError

    def adjust_time(
        self,
        func: Callable[[int], int],
        attr: str = None,
        recursive: bool = True,
    ) -> "Note":
        """Adjust the timing of the note.

        Parameters
        ----------
        func : callable
            The function used to compute the new timing from the old
            timing, i.e., `new_time = func(old_time)`.
        attr : str, optional
            Attribute to adjust. Defaults to adjust all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        """
        if attr is not None and attr != "time":
            raise AttributeError(f"'Note' object has no attribute '{attr}'")

        old_time = self.time
        self.time = func(self.time)
        self.duration = func(old_time + self.duration) - self.time
        return self

    def transpose(self, semitone: int) -> "Note":
        """Transpose the note by a number of semitones.

        Parameters
        ----------
        semitone : int
            Number of semitones to transpose the note. A positive value
            raises the pitch, while a negative value lowers the pitch.

        Returns
        -------
        Object itself.

        """
        self.pitch += semitone
        return self

    def clip(self, lower: int = 0, upper: int = 127) -> "Note":
        """Clip the velocity of the note.

        Parameters
        ----------
        lower : int, default: 0
            Lower bound.
        upper : int, default: 127
            Upper bound.

        Returns
        -------
        Object itself.

        """
        assert upper >= lower, "`upper` must be greater than `lower`."
        if self.velocity > upper:
            self.velocity = upper
        elif self.velocity < lower:
            self.velocity = lower
        return self


class Chord(Base):
    """A container for chords.

    Attributes
    ----------
    time : int
        Start time of the chord, in time steps.
    pitches : list of int
        Note pitches, as MIDI note numbers. Valid values are 0 to 127.
    duration : int
        Duration of the chord, in time steps.
    velocity : int, default: `muspy.DEFAULT_VELOCITY` (64)
        Chord velocity. Valid values are 0 to 127.
    pitches_str : list of str, optional
        Note pitches as strings, useful for distinguishing, e.g., C# and
        Db.

    """

    _attributes = OrderedDict(
        [
            ("time", int),
            ("pitches", int),
            ("duration", int),
            ("velocity", int),
            ("pitches_str", str),
        ]
    )
    _optional_attributes = ["velocity", "pitches_str"]

    def __init__(
        self,
        time: int,
        pitches: List[int],
        duration: int,
        velocity: int = None,
        pitches_str: List[int] = None,
    ):
        self.time = time
        self.pitches = pitches
        self.duration = duration
        self.velocity = velocity if velocity is not None else DEFAULT_VELOCITY
        self.pitches_str = pitches_str

    @property
    def start(self):
        """Start time of the chord."""
        return self.time

    @start.setter
    def start(self, start):
        """Setter for start time."""
        self.time = start

    @property
    def end(self):
        """End time of the chord."""
        return self.time + self.duration

    @end.setter
    def end(self, end):
        """Setter for end time."""
        self.duration = end - self.time

    def _validate(self, attr: str, recursive: bool):
        super()._validate(attr, recursive)
        if attr == "pitches":
            for pitch in self.pitches:
                if pitch < 0 or pitch > 127:
                    raise ValueError(
                        "`pitches` must be a list of integers between 0 to "
                        "127."
                    )
        if attr == "duration" and self.duration < 0:
            raise ValueError("`duration` must be nonnegative.")
        if attr == "velocity" and (self.velocity < 0 or self.velocity > 127):
            raise ValueError("`velocity` must be in between 0 to 127.")

    def _adjust_time(
        self, func: Callable[[int], int], attr: str, recursive: bool
    ):
        raise NotImplementedError

    def adjust_time(
        self,
        func: Callable[[int], int],
        attr: str = None,
        recursive: bool = True,
    ) -> "Chord":
        """Adjust the timing of the chord.

        Parameters
        ----------
        func : callable
            The function used to compute the new timing from the old
            timing, i.e., `new_time = func(old_time)`.
        attr : str, optional
            Attribute to adjust. Defaults to adjust all attributes.
        recursive : bool, default: True
            Whether to apply recursively.

        Returns
        -------
        Object itself.

        """
        if attr is not None and attr != "time":
            raise AttributeError(f"'Note' object has no attribute '{attr}'")

        old_time = self.time
        self.time = func(self.time)
        self.duration = func(old_time + self.duration) - self.time
        return self

    def transpose(self, semitone: int) -> "Chord":
        """Transpose the notes by a number of semitones.

        Parameters
        ----------
        semitone : int
            Number of semitones to transpose the notes. A positive value
            raises the pitches, while a negative value lowers the
            pitches.

        Returns
        -------
        Object itself.

        """
        self.pitches += [pitch + semitone for pitch in self.pitches]
        return self

    def clip(self, lower: int = 0, upper: int = 127) -> "Chord":
        """Clip the velocity of the chord.

        Parameters
        ----------
        lower : int, default: 0
            Lower bound.
        upper : int, default: 127
            Upper bound.

        Returns
        -------
        Object itself.

        """
        assert upper >= lower, "`upper` must be greater than `lower`."
        if self.velocity > upper:
            self.velocity = upper
        elif self.velocity < lower:
            self.velocity = lower
        return self


class Track(ComplexBase):
    """A container for music track.

    Attributes
    ----------
    program : int, default: 0 (Acoustic Grand Piano)
        Program number, according to General MIDI specification [1]_.
        Valid values are 0 to 127.
    is_drum : bool, default: False
        Whether it is a percussion track.
    name : str, optional
        Track name.
    notes : list of :class:`muspy.Note`, default: []
        Musical notes.
    chords : list of :class:`muspy.Chord`, default: []
        Chords.
    annotations : list of :class:`muspy.Annotation`, default: []
        Annotations.
    lyrics : list of :class:`muspy.Lyric`, default: []
        Lyrics.

    Note
    ----
    Indexing a Track object returns the note at a certain index. That
    is, ``track[idx]`` returns ``track.notes[idx]``. Length of a Track
    object is the number of notes. That is, ``len(track)`` returns
    ``len(track.notes)``.

    References
    ----------
    .. [1] https://www.midi.org/specifications/item/gm-level-1-sound-set

    """

    _attributes = OrderedDict(
        [
            ("program", int),
            ("is_drum", bool),
            ("name", str),
            ("notes", Note),
            ("chords", Chord),
            ("lyrics", Lyric),
            ("annotations", Annotation),
        ]
    )
    _optional_attributes = ["name", "notes", "chords", "lyrics", "annotations"]
    _list_attributes = ["notes", "chords", "lyrics", "annotations"]

    def __init__(
        self,
        program: int = 0,
        is_drum: bool = False,
        name: str = None,
        notes: List[Note] = None,
        chords: List[Chord] = None,
        lyrics: List[Lyric] = None,
        annotations: List[Annotation] = None,
    ):
        self.program = program if program is not None else 0
        self.is_drum = is_drum if program is not None else False
        self.name = name
        self.notes = notes if notes is not None else []
        self.chords = chords if chords is not None else []
        self.lyrics = lyrics if lyrics is not None else []
        self.annotations = annotations if annotations is not None else []

    def __len__(self) -> int:
        return len(self.notes)

    def __getitem__(self, key: int) -> Note:
        return self.notes[key]

    def __setitem__(self, key: int, value: Note):
        self.notes[key] = value

    def _validate(self, attr: str, recursive: bool):
        super()._validate(attr, recursive)
        if attr == "program" and self.program < 0 or self.program > 127:
            raise ValueError("`program` must be in between 0 to 127.")

    def get_end_time(self, is_sorted: bool = False) -> int:
        """Return the time of the last event.

        This includes notes, chords, lyrics and annotations.

        Parameters
        ----------
        is_sorted : bool, default: False
            Whether all the list attributes are sorted.

        """

        def _get_end_time(list_, ref_attr="time"):
            if not list_:
                return 0
            if is_sorted:
                return getattr(list_[-1], ref_attr)
            return max(getattr(item, ref_attr) for item in list_)

        return max(
            _get_end_time(self.notes, "end"),
            _get_end_time(self.chords, "end"),
            _get_end_time(self.lyrics),
            _get_end_time(self.annotations),
        )

    def clip(self, lower: int = 0, upper: int = 127) -> "Track":
        """Clip the velocity of each note.

        Parameters
        ----------
        lower : int, default: 0
            Lower bound.
        upper : int, default: 127
            Upper bound.

        Returns
        -------
        Object itself.

        """
        for note in self.notes:
            note.clip(lower, upper)
        return self

    def transpose(self, semitone: int) -> "Track":
        """Transpose the notes by a number of semitones.

        Parameters
        ----------
        semitone : int
            Number of semitones to transpose the notes. A positive value
            raises the pitches, while a negative value lowers the
            pitches.

        Returns
        -------
        Object itself.

        """
        for note in self.notes:
            note.transpose(semitone)
        return self



"""Objective metrics.

This module provides common objective metrics in music generation.
These objective metrics could be used to evaluate a music generation
system by comparing the statistical difference between the training
data and the generated samples.

Functions
---------

- drum_in_pattern_rate
- drum_pattern_consistency
- empty_beat_rate
- empty_measure_rate
- groove_consistency
- n_pitch_classes_used
- n_pitches_used
- pitch_class_entropy
- pitch_entropy
- pitch_in_scale_rate
- pitch_range
- polyphony
- polyphony_rate
- scale_consistency

"""
from .metrics import (
    drum_in_pattern_rate,
    drum_pattern_consistency,
    empty_beat_rate,
    empty_measure_rate,
    groove_consistency,
    n_pitch_classes_used,
    n_pitches_used,
    pitch_class_entropy,
    pitch_entropy,
    pitch_in_scale_rate,
    pitch_range,
    polyphony,
    polyphony_rate,
    scale_consistency,
)

__all__ = [
    "drum_in_pattern_rate",
    "drum_pattern_consistency",
    "empty_beat_rate",
    "empty_measure_rate",
    "groove_consistency",
    "n_pitch_classes_used",
    "n_pitches_used",
    "pitch_class_entropy",
    "pitch_entropy",
    "pitch_in_scale_rate",
    "pitch_range",
    "polyphony",
    "polyphony_rate",
    "scale_consistency",
]



"""JSON, YAML and MusicXML schemas.

This module provide functions for working with MusPy's JSON and YAML
schemas and the MusicXML schema.

Functions
---------

- get_json_schema_path
- get_musicxml_schema_path
- get_yaml_schema_path

Variables
---------
- DEFAULT_SCHEMA_VERSION

"""
from .utils import (
    get_json_schema_path,
    get_musicxml_schema_path,
    get_yaml_schema_path,
    validate_json,
    validate_musicxml,
    validate_yaml,
)
from .version import DEFAULT_SCHEMA_VERSION

__all__ = [
    "DEFAULT_SCHEMA_VERSION",
    "get_json_schema_path",
    "get_musicxml_schema_path",
    "get_yaml_schema_path",
    "validate_json",
    "validate_musicxml",
    "validate_yaml",
]



"""MusicXML input interface."""
import xml.etree.ElementTree as ET
from collections import OrderedDict
from functools import reduce
from operator import attrgetter
from pathlib import Path
from typing import Dict, List, Optional, Tuple, TypeVar, Union
from xml.etree.ElementTree import Element
from zipfile import ZipFile

from ..classes import (
    KeySignature,
    Lyric,
    Metadata,
    Note,
    Tempo,
    TimeSignature,
    Track,
)
from ..music import Music
from ..utils import CIRCLE_OF_FIFTHS, MODE_CENTERS, NOTE_MAP, NOTE_TYPE_MAP

T = TypeVar("T")


class MusicXMLError(Exception):
    """An error class for MusicXML related exceptions."""


def _gcd(a: int, b: int) -> int:
    """Return greatest common divisor using Euclid's Algorithm.

    Code copied from https://stackoverflow.com/a/147539.

    """
    while b:
        a, b = b, a % b
    return a


def _lcm_two_args(a: int, b: int) -> int:
    """Return least common multiple.

    Code copied from https://stackoverflow.com/a/147539.

    """
    return a * b // _gcd(a, b)


def _lcm(*args: int) -> int:
    """Return lcm of args.

    Code copied from https://stackoverflow.com/a/147539.

    """
    return reduce(_lcm_two_args, args)  # type: ignore


def _get_text(
    element: Element,
    path: str,
    default: T = None,
    remove_newlines: bool = False,
) -> Union[str, T]:
    """Return the text of the first matching element."""
    elem = element.find(path)
    if elem is not None and elem.text is not None:
        if remove_newlines:
            return " ".join(elem.text.splitlines())
        return elem.text
    return default  # type: ignore


def _get_required(element: Element, path: str) -> Element:
    """Return a required element; raise ValueError if not found."""
    elem = element.find(path)
    if elem is None:
        raise MusicXMLError("Element `{}` is required.".format(path))
    return elem


def _get_required_attr(element: Element, attr: str) -> str:
    """Return a required attribute; raise MusicXMLError if not found."""
    attribute = element.get(attr)
    if attribute is None:
        raise MusicXMLError("Attribute '{}' is required for an element ")
    return attribute


def _get_required_text(
    element: Element, path: str, remove_newlines: bool = False
) -> str:
    """Return a required text; raise MusicXMLError if not found."""
    elem = element.find(path)
    if elem is None:
        raise MusicXMLError(
            "Child element '{}' is required for an element '{}'."
            "".format(path, element.tag)
        )
    if elem.text is None:
        raise MusicXMLError(
            "Text content '{}' of an element '{}' must not be empty."
            "".format(path, element.tag)
        )
    if remove_newlines:
        return " ".join(elem.text.splitlines())
    return elem.text


def parse_metronome_elem(elem: Element) -> Optional[float]:
    """Return a qpm value parsed from a metronome element."""
    beat_unit = _get_text(elem, "beat-unit")
    if beat_unit is not None:
        per_minute = _get_text(elem, "per-minute")
        if per_minute is not None and beat_unit in NOTE_TYPE_MAP:
            qpm = NOTE_TYPE_MAP[beat_unit] * float(per_minute)
            if elem.find("beat-unit-dot") is not None:
                qpm *= 1.5
            return qpm
    return None


def parse_key_elem(elem: Element) -> Dict:
    """Return a dictionary with data parsed from a key element."""
    mode = _get_text(elem, "mode", "major")
    fifths = int(_get_required_text(elem, "fifths"))
    if mode is None:
        return {"fifths": fifths}
    idx = MODE_CENTERS[mode] + fifths
    if idx < 0 or idx > 20:
        return {"fifths": fifths, "mode": mode}
    root, root_str = CIRCLE_OF_FIFTHS[MODE_CENTERS[mode] + fifths]
    return {"root": root, "mode": mode, "fifths": fifths, "root_str": root_str}


def parse_pitch_elem(elem: Element) -> Tuple[int, str]:
    """Return a (pitch, pitch_str) tuple parsed from a pitch element."""
    step = _get_required_text(elem, "step")
    octave = int(_get_required_text(elem, "octave"))
    alter = int(_get_text(elem, "alter", 0))
    pitch = 12 * (octave + 1) + NOTE_MAP[step] + alter
    if alter > 0:
        pitch_str = step + "#" * alter + str(octave)
    elif alter < 0:
        pitch_str = step + "b" * (-alter) + str(octave)
    else:
        pitch_str = step + str(octave)
    return pitch, pitch_str


def parse_part_elem(
    part_elem: Element, resolution: int, instrument_info: dict
) -> dict:
    """Return a dictionary with data parsed from a part element."""
    # Initialize lists and placeholders
    tempos: List[Tempo] = []
    key_signatures: List[KeySignature] = []
    time_signatures: List[TimeSignature] = []
    notes: Dict[str, List[Note]] = {
        instrument_id: [] for instrument_id in instrument_info
    }
    lyrics: List[Lyric] = []
    ties: Dict[Tuple[str, int], int] = {}

    # Initialize variables
    time = 0
    velocity = 64
    division = 1
    default_instrument_id = next(iter(instrument_info))
    transpose_semitone = 0
    transpose_octave = 0
    # Repeats
    is_repeat = 0
    last_repeat = 0
    count_repeat = 1
    count_ending = 1
    # Coda, tocoda, dacapo, segno, dalsegno, fine
    is_after_jump = False
    is_fine = False
    is_dacapo = False
    is_dalsegno = False
    is_segno = False
    is_segno_found = False
    is_tocoda = False
    is_coda = False
    is_coda_found = False

    # Iterate over all elements
    measure_idx = 0
    measure_elems = list(part_elem.findall("measure"))
    while measure_idx < len(measure_elems):

        # Get the measure element
        measure_elem = measure_elems[measure_idx]

        # Initialize position
        position = 0
        last_note_position = None

        # Look for segno
        if is_dalsegno and not is_segno_found:
            # Segno
            for sound_elem in measure_elem.findall("sound"):
                if sound_elem.get("segno") is not None:
                    is_segno = True
            for sound_elem in measure_elem.findall("direction/sound"):
                if sound_elem.get("segno") is not None:
                    is_segno = True

            # Skip if not segno
            if not is_segno:
                measure_idx += 1
                continue

            is_segno_found = True

        # Look for coda
        if is_tocoda and not is_coda_found:
            # Coda
            for sound_elem in measure_elem.findall("sound"):
                if sound_elem.get("coda") is not None:
                    is_coda = True
            for sound_elem in measure_elem.findall("direction/sound"):
                if sound_elem.get("coda") is not None:
                    is_coda = True

            # Skip if not coda
            if not is_coda:
                measure_idx += 1
                continue

            is_coda_found = True

        # Sound element
        for sound_elem in measure_elem.findall("sound"):
            if is_after_jump:
                # Tocoda
                if sound_elem.get("tocoda") is not None:
                    is_tocoda = True

                # Fine
                if sound_elem.get("fine") is not None:
                    is_fine = True
            else:
                # Dacapo
                if sound_elem.get("dacapo") is not None:
                    is_dacapo = True

                # Daselgno
                if sound_elem.get("dalsegno") is not None:
                    is_dalsegno = True

        # Sound elements under direction elements
        for sound_elem in measure_elem.findall("direction/sound"):
            if is_after_jump:
                # Tocoda
                if sound_elem.get("tocoda") is not None:
                    is_tocoda = True

                # Fine
                if sound_elem.get("fine") is not None:
                    is_fine = True
            else:
                # Dacapo
                if sound_elem.get("dacapo") is not None:
                    is_dacapo = True

                # Daselgno
                if sound_elem.get("dalsegno") is not None:
                    is_dalsegno = True

        # Barline elements
        for barline_elem in measure_elem.findall("barline"):
            # Repeat elements
            repeat_elem = barline_elem.find("repeat")
            if repeat_elem is not None:
                direction = _get_required_attr(repeat_elem, "direction")
                if direction == "forward":
                    last_repeat = measure_idx
                elif direction == "backward":
                    # Get after-jump infomation
                    after_jump_attr = repeat_elem.get("after-jump")
                    if after_jump_attr is None or after_jump_attr == "no":
                        after_jump = False
                    else:
                        after_jump = True
                    if not is_after_jump or (is_after_jump and after_jump):
                        # Get repeat-times infomation
                        repeat_times_attr = repeat_elem.get("times")
                        if repeat_times_attr is None:
                            repeat_times = 2
                        else:
                            repeat_times = int(repeat_times_attr)
                        # Check if repeat times has reached
                        if count_repeat < repeat_times:
                            count_repeat += 1
                            count_ending += 1
                            is_repeat = True
                        else:
                            count_repeat = 1
                            count_ending = 1
                else:
                    raise MusicXMLError(
                        "Unknown direction for a `repeat` element : "
                        f"{direction}"
                    )
            # Ending elements
            ending_elem = barline_elem.find("ending")
            if ending_elem is not None:
                ending_num_attr = _get_required_attr(ending_elem, "number")
                if ending_num_attr:
                    ending_num = [
                        int(num) for num in ending_num_attr.split(",")
                    ]
                # Skip the current measure if not the correct ending
                if not is_repeat and count_ending not in ending_num:
                    measure_idx += 1
                    continue

        # Iterating over all elements in the current measure
        for elem in measure_elem:
            # Attributes elements
            if elem.tag == "attributes":
                # Division elements
                division_elem = elem.find("divisions")
                if (
                    division_elem is not None
                    and division_elem.text is not None
                ):
                    division = int(division_elem.text)

                # Transpose elements
                transpose_elem = elem.find("transpose")
                if transpose_elem is not None:
                    transpose_semitone = int(
                        _get_required_text(transpose_elem, "chromatic")
                    )
                    octave_change = _get_text(transpose_elem, "octave-change")
                    if octave_change is not None:
                        transpose_octave = int(octave_change)

                # Time signatures
                time_elem = elem.find("time")
                if time_elem is not None:
                    # Numerator
                    beats = _get_required_text(time_elem, "beats")
                    if "+" in beats:
                        numerator = sum(int(beat) for beat in beats.split("+"))
                    else:
                        numerator = int(beats)

                    # Denominator
                    beat_type = _get_required_text(time_elem, "beat-type")
                    if "+" in beat_type:
                        raise RuntimeError(
                            "Compound time signatures with separate fractions "
                            "are not supported."
                        )
                    denominator = int(beat_type)
                    time_signatures.append(
                        TimeSignature(
                            time=time + position,
                            numerator=numerator,
                            denominator=denominator,
                        )
                    )

                # Key elements
                key_elem = elem.find("key")
                if key_elem is not None:
                    parsed_key = parse_key_elem(key_elem)
                    if parsed_key is not None:
                        key_signatures.append(
                            KeySignature(
                                time=time + position,
                                root=parsed_key.get("root"),
                                mode=parsed_key.get("mode"),
                                root_str=parsed_key.get("root_str"),
                            )
                        )

            # Sound element
            elif elem.tag == "sound":
                # Tempo elements
                tempo = elem.get("tempo")
                if tempo is not None:
                    tempos.append(Tempo(time + position, float(tempo)))

                # Dynamics elements
                dynamics = elem.get("dynamics")
                if dynamics is not None:
                    velocity = round(float(dynamics))

            # Direction elements
            elif elem.tag == "direction":
                # TODO: Handle symbolic dynamics and tempo

                tempo_set = False

                # Sound elements
                sound_elem_ = elem.find("sound")
                if sound_elem_ is not None:
                    # Tempo directions
                    tempo = sound_elem_.get("tempo")
                    if tempo is not None:
                        tempos.append(
                            Tempo(time=time + position, qpm=float(tempo))
                        )
                        tempo_set = True

                    # Dynamic directions
                    dynamics = sound_elem_.get("dynamics")
                    if dynamics is not None:
                        velocity = round(float(dynamics))

                # Metronome elements
                if not tempo_set:
                    metronome_elem = elem.find("direction-type/metronome")
                    if metronome_elem is not None:
                        qpm = parse_metronome_elem(metronome_elem)
                        if qpm is not None:
                            tempos.append(Tempo(time=time + position, qpm=qpm))

            # Note elements
            elif elem.tag == "note":
                # TODO: Handle voice information

                # Rest elements
                rest_elem = elem.find("rest")
                if rest_elem is not None:
                    # Move time position forward if it is a rest
                    duration = int(_get_required_text(elem, "duration"))
                    position += round(duration * resolution / division)
                    continue

                # Cue notes
                if elem.find("cue") is not None:
                    continue

                # Unpitched notes
                # TODO: Handle unpitched notes
                unpitched_elem = elem.find("unpitched")
                if unpitched_elem is not None:
                    continue

                # Chord elements
                if elem.find("chord") is not None:
                    # Move time position backward if it is in a chord
                    if last_note_position is not None:
                        position = last_note_position

                # Compute pitch number
                pitch, pitch_str = parse_pitch_elem(
                    _get_required(elem, "pitch")
                )
                pitch += 12 * transpose_octave + transpose_semitone

                # Get instrument information
                instrument_elem = elem.find("instrument")
                if instrument_elem is not None:
                    instrument_id = _get_required_text(instrument_elem, "id")
                    if instrument_id not in instrument_info:
                        raise MusicXMLError(
                            "ID of an 'instrument' element must be predefined "
                            "in a 'score-instrument' element."
                        )
                else:
                    instrument_id = default_instrument_id

                # Grace notes
                grace_elem = elem.find("grace")
                if grace_elem is not None:
                    note_type = _get_required_text(elem, "type")
                    notes[instrument_id].append(
                        Note(
                            time=time + position,
                            pitch=pitch,
                            duration=round(
                                NOTE_TYPE_MAP[note_type] * resolution
                            ),
                            velocity=velocity,
                            pitch_str=pitch_str,
                        )
                    )
                    continue

                # Get duration
                # TODO: Should we look for a duration or type element?
                duration = int(_get_required_text(elem, "duration"))

                # Check if it is a tied note
                # TODO: Should we look for a tie or tied element?
                is_outgoing_tie = False
                for tie_elem in elem.findall("tie"):
                    if tie_elem.get("type") == "start":
                        is_outgoing_tie = True

                # Check if it is an incoming tied note
                note_key = (instrument_id, pitch)
                if note_key in ties:
                    note_idx = ties[note_key]
                    notes[instrument_id][note_idx].duration += round(
                        duration * resolution / division
                    )

                    if is_outgoing_tie:
                        ties[note_key] = note_idx
                    else:
                        del ties[note_key]

                else:
                    # Create a new note and append it to the note list
                    notes[instrument_id].append(
                        Note(
                            time=time + position,
                            pitch=pitch,
                            duration=round(duration * resolution / division),
                            velocity=velocity,
                            pitch_str=pitch_str,
                        )
                    )

                    if is_outgoing_tie:
                        ties[note_key] = len(notes[instrument_id]) - 1

                # Lyrics
                lyric_elem = elem.find("lyric")
                if lyric_elem is not None:
                    lyric_text = _get_required_text(lyric_elem, "text")
                    syllabic_elem = lyric_elem.find("syllabic")
                    if syllabic_elem is not None:
                        if syllabic_elem.text == "begin":
                            lyric_text += "-"
                        elif syllabic_elem.text == "middle":
                            lyric_text = "-" + lyric_text + "-"
                        elif syllabic_elem.text == "end":
                            lyric_text = "-" + lyric_text
                    lyrics.append(
                        Lyric(time=time + position, lyric=lyric_text)
                    )

                # Move time position forward if it is not in chord
                last_note_position = position
                position += round(duration * resolution / division)

            # Forward elements
            elif elem.tag == "forward":
                duration = int(_get_required_text(elem, "duration"))
                position += round(duration * resolution / division)

            # Backup elements
            elif elem.tag == "backup":
                duration = int(_get_required_text(elem, "duration"))
                position -= round(duration * resolution / division)

        time += position

        if is_after_jump and is_fine:
            break

        if not is_after_jump and (is_dacapo or is_dalsegno):
            measure_idx = 0
            is_after_jump = True
        elif is_repeat:
            is_repeat = False
            measure_idx = last_repeat
        else:
            measure_idx += 1

    # Sort notes
    for instrument_notes in notes.values():
        instrument_notes.sort(
            key=attrgetter("time", "pitch", "duration", "velocity")
        )

    # Sort tempos, key signatures, time signatures and lyrics
    tempos.sort(key=attrgetter("time"))
    key_signatures.sort(key=attrgetter("time"))
    time_signatures.sort(key=attrgetter("time"))
    lyrics.sort(key=attrgetter("time"))

    return {
        "tempos": tempos,
        "key_signatures": key_signatures,
        "time_signatures": time_signatures,
        "notes": notes,
        "lyrics": lyrics,
    }


def parse_metadata(root: Element) -> Metadata:
    """Return a Metadata object parsed from a MusicXML file."""
    # Title is usually stored in movement-title. See
    # https://www.musicxml.com/tutorial/file-structure/score-header-entity/
    title = _get_text(root, "movement-title", remove_newlines=True)
    if not title:
        title = _get_text(root, "work/work-title", remove_newlines=True)

    # Creators and copyrights
    creators = []
    copyrights = []

    identification_elem = root.find("identification")
    if identification_elem is not None:
        for creator_elem in identification_elem.findall("creator"):
            if creator_elem.text:
                creators.append(creator_elem.text)
        for right_elem in identification_elem.findall("rights"):
            if right_elem.text:
                copyrights.append(right_elem.text)

    return Metadata(
        title=title,
        creators=creators,
        copyright=" ".join(copyrights) if copyrights else None,
        source_format="musicxml",
    )


def _get_root(path: Union[str, Path], compressed: bool = None):
    """Return root of the element tree."""
    if compressed is None:
        compressed = str(path).endswith(".mxl")

    if not compressed:
        tree = ET.parse(str(path))
        return tree.getroot()

    # Find out the main MusicXML file in the compressed ZIP archive
    # according to the official tutorial (see
    # https://www.musicxml.com/tutorial/compressed-mxl-files/).
    zip_file = ZipFile(str(path))
    if "META-INF/container.xml" not in zip_file.namelist():
        raise MusicXMLError("Container file ('container.xml') not found.")
    container = ET.fromstring(zip_file.read("META-INF/container.xml"))
    rootfile = container.find("rootfiles/rootfile")
    if rootfile is None:
        raise MusicXMLError(
            "Element 'rootfile' tag not found in the container file "
            "('container.xml')."
        )
    filename = _get_required_attr(rootfile, "full-path")
    return ET.fromstring(zip_file.read(filename))


def _get_divisions(root: Element):
    """Return a list of divisions."""
    divisions = []
    for division_elem in root.findall("part/measure/attributes/divisions"):
        if division_elem.text is None:
            continue
        if not float(division_elem.text).is_integer():
            raise MusicXMLError(
                "Noninteger 'division' values are not supported."
            )
        divisions.append(int(division_elem.text))
    return divisions


def parse_score_part_elem(elem: Element) -> Tuple[str, OrderedDict]:
    """Return part information parsed from a score part element."""
    # Part ID
    part_id = _get_required_attr(elem, "id")

    # Part name
    part_name = _get_text(elem, "part-name", remove_newlines=True)

    # Instruments
    part_info: OrderedDict = OrderedDict()
    for score_instrument_elem in elem.findall("score-instrument"):
        instrument_id = _get_required_attr(score_instrument_elem, "id")
        part_info[instrument_id] = OrderedDict()
        part_info[instrument_id]["name"] = _get_text(
            score_instrument_elem,
            "instrument-name",
            part_name,
            remove_newlines=True,
        )
    for midi_instrument_elem in elem.findall("midi-instrument"):
        instrument_id = _get_required_attr(midi_instrument_elem, "id")
        if instrument_id not in part_info:
            if instrument_id == part_id:
                instrument_id = ""
                part_info[""] = {"name": part_name}
            else:
                raise MusicXMLError(
                    "ID of a 'midi-instrument' element must be predefined "
                    "in a 'score-instrument' element."
                )
        part_info[instrument_id]["program"] = int(
            _get_text(midi_instrument_elem, "midi-program", 0)
        )
        part_info[instrument_id]["is_drum"] = (
            int(_get_text(midi_instrument_elem, "midi-channel", 0)) == 10
        )
    if not part_info:
        part_info[""] = {"name": part_name}
    for value in part_info.values():
        if "program" not in value:
            value["program"] = 0
        if "is_drum" not in value:
            value["is_drum"] = False
    return part_id, part_info


def read_musicxml(
    path: Union[str, Path], resolution: int = None, compressed: bool = None,
) -> Music:
    """Read a MusicXML file into a Music object.

    Parameters
    ----------
    path : str or Path
        Path to the MusicXML file to read.
    resolution : int, optional
        Time steps per quarter note. Defaults to the least common
        multiple of all divisions.
    compressed : bool, optional
        Whether it is a compressed MusicXML file. Defaults to infer
        from the filename.

    Returns
    -------
    :class:`muspy.Music`
        Converted Music object.

    Notes
    -----
    Grace notes and unpitched notes are not supported.

    """
    # Get element tree root
    root = _get_root(path, compressed)

    if root.tag == "score-timewise":
        raise ValueError("MusicXML file with timewise type is not supported.")

    # Meta data
    metadata = parse_metadata(root)
    metadata.source_filename = Path(path).name

    # Set resolution to the least common multiple of all divisions
    if resolution is None:
        divisions = _get_divisions(root)
        resolution = _lcm(*divisions) if divisions else 1

    # Part information
    part_info: OrderedDict = OrderedDict()
    for part_elem in root.findall("part-list/score-part"):
        part_id, info = parse_score_part_elem(part_elem)
        part_info[part_id] = info

    if root.find("part") is None:
        return Music(metadata=metadata, resolution=resolution)

    # Initialize lists
    tempos: List[Tempo] = []
    key_signatures: List[KeySignature] = []
    time_signatures: List[TimeSignature] = []
    tracks: List[Track] = []

    # Raise an error if part-list information is missing for a
    # multi-part piece
    if not part_info:
        if len(root.findall("part")) > 1:
            raise MusicXMLError(
                "Part-list information is required for a multi-part piece."
            )
        part_elem = _get_required(root, "part")
        instrument_info = {"": {"program": 0, "is_drum": False}}
        part = parse_part_elem(part_elem, resolution, instrument_info)

    else:
        # Iterate over all parts and measures
        for part_elem in root.findall("part"):
            part_id = part_elem.get("id")  # type: ignore
            if part_id is None:
                if len(root.findall("part")) > 1:
                    continue
                part_id = next(iter(part_info))
            if part_id not in part_info:
                continue

            # Parse part
            part = parse_part_elem(part_elem, resolution, part_info[part_id])

            # Extend lists
            tempos.extend(part["tempos"])
            key_signatures.extend(part["key_signatures"])
            time_signatures.extend(part["time_signatures"])
            for instrument_id, notes in part["notes"].items():
                track = Track(
                    program=part_info[part_id][instrument_id]["program"],
                    is_drum=part_info[part_id][instrument_id]["is_drum"],
                    name=part_info[part_id][instrument_id]["name"],
                    notes=notes,
                    lyrics=part["lyrics"],
                )
                tracks.append(track)

    # Sort tempos, key signatures and time signatures
    tempos.sort(key=attrgetter("time"))
    key_signatures.sort(key=attrgetter("time"))
    time_signatures.sort(key=attrgetter("time"))

    return Music(
        metadata=metadata,
        resolution=resolution,
        tempos=tempos,
        key_signatures=key_signatures,
        time_signatures=time_signatures,
        tracks=tracks,
    )



"""External dependencies.

This module provides functions for working with external dependencies.

Functions
---------

- download_bravura_font
- download_musescore_soundfont
- get_bravura_font_dir
- get_bravura_font_path
- get_musescore_soundfont_dir
- get_musescore_soundfont_path

"""
import urllib.request
from pathlib import Path

__all__ = [
    "download_bravura_font",
    "download_musescore_soundfont",
    "get_bravura_font_dir",
    "get_bravura_font_path",
    "get_musescore_soundfont_dir",
    "get_musescore_soundfont_path",
]


def get_bravura_font_dir() -> Path:
    """Return path to the directory of the Bravura font."""
    return Path.home() / ".muspy/bravura"


def get_bravura_font_path() -> Path:
    """Return path to the Bravura font."""
    return get_bravura_font_dir() / "Bravura.otf"


def download_bravura_font(overwrite: bool = False):
    """Download the Bravura font.

    Parameters
    ----------
    overwrite : bool, default: False
        Whether to overwrite an existing file.

    """
    if not overwrite and get_bravura_font_path().is_file():
        print("Skip downloading as the Bravura font is found.")
        return

    # Make sure the directory exists
    get_bravura_font_dir().mkdir(parents=True, exist_ok=True)

    # Download the font
    print("Start downloading Bravura font.")
    prefix = "https://github.com/steinbergmedia/bravura/raw/master/"
    urllib.request.urlretrieve(
        prefix + "redist/otf/Bravura.otf", get_bravura_font_path()
    )
    print(
        "Bravura font has successfully been downloaded to : "
        "{}.".format(get_musescore_soundfont_dir())
    )

    # Download the license
    urllib.request.urlretrieve(
        prefix + "LICENSE.txt", get_bravura_font_dir() / "LICENSE.txt"
    )


def get_musescore_soundfont_dir() -> Path:
    """Return path to the MuseScore General soundfont directory."""
    return Path.home() / ".muspy/musescore-general"


def get_musescore_soundfont_path() -> Path:
    """Return path to the MuseScore General soundfont."""
    return get_musescore_soundfont_dir() / "MuseScore_General.sf3"


def download_musescore_soundfont(overwrite: bool = False):
    """Download the MuseScore General soundfont.

    Parameters
    ----------
    overwrite : bool, default: False
        Whether to overwrite an existing file.

    """
    if not overwrite and get_musescore_soundfont_path().is_file():
        print("Skip downloading as the MuseScore General soundfont is found.")
        return

    # Make sure the directory exists
    get_musescore_soundfont_dir().mkdir(parents=True, exist_ok=True)

    # Download the soundfont
    print("Start downloading MuseScore General soundfont.")
    prefix = "ftp://ftp.osuosl.org/pub/musescore/soundfont/MuseScore_General/"
    urllib.request.urlretrieve(
        prefix + "MuseScore_General.sf3", get_musescore_soundfont_path()
    )
    print(
        "MuseScore General soundfont has successfully been downloaded to : "
        "{}.".format(get_musescore_soundfont_dir())
    )

    # Download the license
    urllib.request.urlretrieve(
        prefix + "MuseScore_General_License.md",
        get_musescore_soundfont_dir() / "MuseScore_General_License.md",
    )
